#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <strings.h>
#include <syslog.h>
#include <inttypes.h>
#include <pwd.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "mhvtl_list.h"
#include "be_byteshift.h"
#include "vtl_common.h"
#include "mhvtl_scsi.h"
#include "q.h"
#include "logging.h"
#include "vtllib.h"
#include "vtltape.h"
#include "spc.h"
#include "ssc.h"
#include "mhvtl_log.h"
#include "mode.h"

int verbose = 0;
int debug = 0;
long my_id;
int lib_id;
int cdbonly = 0;

char mhvtl_driver_name[] = "vtl_remove_dev";

static void usage(char *progname) 
{
	return;
}
static int cdev;
static struct mhvtl_ctl ctl;

int main(int argc, char *argv[])
{

	char *progname = argv[0];
	char *name = "mhvtl";
	unsigned minor = 0;

	memset(&ctl, 0, sizeof(struct mhvtl_ctl));
	if (argc < 3) {
		usage(argv[0]);
		printf("  -- Not enough parameters --\n");
		exit(1);
	}

	while (argc > 0) {
		if (argv[0][0] == '-') {
			switch (argv[0][1]) {
			case 'q':
				if (argc > 1)
					my_id = atoi(argv[1]);
				break;
			case 'l':
				if (argc > 1)
					lib_id = atoi(argv[1]);
				 break;
			default:
				usage(progname);
				printf("    Unknown option %c\n", argv[0][1]);
				exit(1);
				break;
			}
		}
		argv++;
		argc--;
	}

	if (my_id <= 0 || my_id > MAXPRIOR) {
		usage(progname);
		if (my_id == 0)
			puts("    -q must be specified\n");
		else
			printf("    -q value out of range [1 - %d]\n",
				MAXPRIOR);
		exit(1);
	}
	
	minor = my_id;	/* Minor == Message Queue priority */

	ctl.channel = 0;
	ctl.id = lib_id;
	if (my_id > 20)
		ctl.lun = my_id - ((lib_id*60) - 40);
	else
		ctl.lun = 0;

	cdev = chrdev_open(name, minor);
	if (cdev == -1) {
		printf("Could not open /dev/%s%u: %s", name, minor,
						strerror(errno));
		fflush(NULL);
		exit(1);
	}
	ioctl(cdev, VTL_UNBUSY_LU, &ctl);
	ioctl(cdev, VTL_REMOVE_LU, &ctl);

	close(cdev);
	
	exit(0);
}

