/*
 * This daemon is the SCSI SMC target (Medium Changer) portion of the
 * vtl package.
 *
 * The vtl package consists of:
 *   a kernel module (vlt.ko) - Currently on 2.6.x Linux kernel support.
 *   SCSI target daemons for both SMC and SSC devices.
 *
 * Copyright (C) 2005 - 2009 Mark Harvey       markh794@gmail.com
 *                                          mark.harvey at nutanix.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * See comments in vtltape.c for a more complete version release...
 *
 * 0.14 13 Feb 2008
 *	Since ability to define device serial number, increased ver from
 *	0.12 to 0.14
 *
 * v0.12 -> Forked into 'stable' (0.12) and 'devel' (0.13).
 *          My current thinking : This is a dead end anyway.
 *          An iSCSI target done in user-space is now my perferred solution.
 *          This means I don't have to do any kernel level drivers
 *          and leaverage the hosts native iSCSI initiator.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <string.h>
#include <syslog.h>
#include <ctype.h>
#include <inttypes.h>
#include <signal.h>
#include <pwd.h>
#include <math.h>
#include "vtl_common.h"
#include "mhvtl_scsi.h"
#include "mhvtl_list.h"
#include "q.h"
#include "logging.h"
#include "vtltape.h"
#include "vtllib.h"
#include "spc.h"
#include "smc.h"
#include "mode.h"
#include "be_byteshift.h"
#include "mhvtl_log.h"
#undef __USE_FILE_OFFSET64
#include <dirent.h>
#include <sys/types.h>
#include <attr/xattr.h>

/* Strict support for STK L700 */
#define STK 1
#define MAX_IOSZ (256 * 1024)

char mhvtl_driver_name[] = "vtllibrary";
long my_id = 0;
static char *cont_name = NULL;
unsigned int no_of_drives;
static char home_directory[1024];
static char conf_directory[1024];
static char cart_home[1024];
static int gslots;
static bool is_oem=false;
static bool archive = false;
static char *oem_vendor_id = NULL; 
static bool tdreq_unload_tape = false;

#define CAP_CLOSED	1
#define CAP_OPEN	0
#define OPERATOR	1
#define ROBOT_ARM	0

#define SMC_BUF_SIZE 1024 * 1024 /* Default size of buffer */

#define DEVICE_CONF 	"/device.conf"
#define VENDOR_ID	"STK"
#define OEM_PRODUCT_ID	"DR_L700"
#define PRODUCT_ID	"L700"

// [12 TB LTO 8 Data Cartridge (xxxxxxL8)]
// [Native capacity - 12 TB (10.91 TiB)]
// [Compressed capacity - 30 TB (27.93 TiB)]
static char density[] = "LTO8";

#define LIBCONTENTS "/library_contents."

int verbose = 0;
int reinstated = 0;
int debug = 0;
int cdbonly = 0;
static uint8_t sam_status = 0;	/* Non-zero if Sense-data is valid */
static long backoff;		/* Backoff value for polling char device */
/*
 * Serial base for all devices and carts
 */
static char g_serial_base[32];	/* Serial base for all devices and carts */
/*
 * For salting serial numbers of tape drives, library and 
 *  cartridges.
 */
static char g_repl_salt[3];	/* Replication salt */
extern int current_state;	/* Scope, Global -> Last status sent to fifo */

/*
 * Logical Unit structure 
 */
struct lu_phy_attr lunit;
pthread_mutex_t	import_lock;
struct serial_list import_serial_list;

struct serials *sortedmerge(struct serials *a, struct serials *b);
void split(struct serials *source, struct serials **frontref,
	   struct serials **backref);

int tape_size=0;
int max_s_slot;

struct smc_priv smc_slots;

static int create_carts(struct q_msg *msg);
static char *__get_base36(int base10, char *alphanum);
static int create_missing_carts(const char *drsnap_recovery_marker);
/*
 * vtl --show stuff
 */
static int push_show = 1;
static char g_f[32] = { 0 };
static FILE *g_fp = NULL;

struct s_info *add_new_slot(struct lu_phy_attr *lu);

static void usage(char *progname)
{
  return;
}

struct MAM *init_mam(struct MAM *mamp, const char *pcl)
{
  memset((uint8_t *)mamp, 0, sizeof(*mamp));
  
  mamp->tape_fmt_version = TAPE_FMT_VERSION;
  mamp->mam_fmt_version = MAM_VERSION;
  // tape_size in MBs
  put_unaligned_be64((uint64_t)tape_size * 1048576, &mamp->max_capacity); // In bytes
  put_unaligned_be64((uint64_t)tape_size * 1048576, &mamp->remaining_capacity); // In bytes
  put_unaligned_be64(sizeof(mamp->pad), &mamp->MAMSpaceRemaining);
  
  memcpy(&mamp->MediumManufacturer, "linuxVTL", 8);
  memcpy(&mamp->ApplicationVendor, "vtl-1.7 ", 8);
  sprintf((char *)mamp->ApplicationVersion, "%d", TAPE_FMT_VERSION);

  mamp->MediumType = MEDIA_TYPE_DATA;       /* Normal data cart */
  set_media_params(&mam, density); // "LTO8" - 12 TBs
  
  snprintf((char *)mamp->MediumSerialNumber, sizeof(mamp->MediumSerialNumber),
	   "%s_%d", pcl, (int)time(NULL));

  snprintf((char *)mamp->MediumManufactureDate, sizeof(mamp->MediumManufactureDate),
	   "%d", (int)time(NULL));
  snprintf((char *)mamp->Barcode, sizeof(mamp->Barcode), "%-31s", pcl);
  
  return mamp;
}

static int create_missing_carts(const char *drsnap_recovery_marker)
{
  char tape_base[10];
  int no_of_tapes = 0;
  FILE *fp = NULL;
  int rval = -1;
  char line[1024];  

  memset(tape_base, 0, ARRAY_SIZE(tape_base));
  memset(line, 0, ARRAY_SIZE(line));

  if ((fp = fopen(drsnap_recovery_marker, "r+"))) {

    if (fgets(line, ARRAY_SIZE(line), fp)) {
      char alphanum[4];
      int i = 0;
      unsigned char sam_stat = 0;
      char spcl[10];
      char *str = NULL, *saveptr = NULL;

      /* tape base token */
      str = strtok_r(line, " ", &saveptr);
      if (str)
	strncpy(tape_base, str, sizeof(tape_base)-1);
      
      /* number of carts token */
      str = strtok_r(NULL, " ", &saveptr);
      if (str)
	no_of_tapes = atoi(str);

      MHVTL_LOG("drsnap marker file's contents: tape_base=%s no_of_tapes=%d", 
		tape_base, no_of_tapes);
      
      rval = 0;

      // create missing carts
      for (i = 1; i<= no_of_tapes; i++){
	int rc = 0, err = 0;

	__get_base36(i, alphanum);

	snprintf(spcl, sizeof(spcl), "%s%s", tape_base, alphanum);

	(void) init_mam(&mam, spcl);

	/*
	 * create tape will do nothing if cart already exists
	 *
	 * 0: the new PCL was successfully created.
	 * 1/2: cart create failed.
	 */
	rc = create_tape(spcl, &mam, &sam_stat, cart_home, &err);
	if (rc) {
	  MHVTL_ERR("Failed to create a cartridge with Barcode: %s", spcl);
	  rval++;
	}
      }
    }
  } else {
    MHVTL_ERR("Failed to open Disaster recovery marker file: %s", drsnap_recovery_marker);
  }

  if (fp) {
    fclose(fp);
    fp = NULL;
  }

  return rval;
}

/* Add salt to a serial number base */
static char *replsalt(char *serial)
{
	static char ser[32];
	int i;
	if (!g_repl_salt[0]) return serial;
	memset(ser, 0, 32);
	memcpy(ser, serial, strlen(serial));
	if (g_repl_salt[0] == '0' && g_repl_salt[1] == '0') {
		return ser;
	}
	for (i = 4; i < 6; i++) {
		ser[i] = (serial[i] + g_repl_salt[i-4]) / 2;
		if (ser[i] > 57 && ser[i] < 65) {
			ser[i] = 'A' + (ser[i] - 57);
		}
	}
	return ser;
}

#ifndef Solaris
 int ioctl(int, int, void *);
#endif

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"

struct device_type_template smc_template = {
	.ops	= {
		/* 0x00 -> 0x0f */
		{spc_tur,},
		{smc_rezero,},
		{spc_illegal_op,},
		{spc_request_sense,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{smc_initialize_element_status,},

		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},

		/* 0x10 -> 0x1f */
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_inquiry,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_mode_select,},
		{spc_reserve,},
		{spc_release,},

		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_mode_sense,},
		{smc_open_close_import_export_element,},
		{spc_recv_diagnostics,},
		{spc_send_diagnostics,},
		{smc_allow_removal,},
		{spc_illegal_op,},

		/* 0x20 -> 0x2f */
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},

		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{smc_position_to_element,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},

		[0x30 ... 0x3f] = {spc_illegal_op,},

		/* 0x40 -> 0x4f */
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},

		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_log_select,},
		{smc_log_sense,},
		{spc_illegal_op,},
		{spc_illegal_op,},

		/* 0x50 -> 0x5f */
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_mode_select,},
		{spc_reserve,},
		{spc_release,},

		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_mode_sense,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},

		[0x60 ... 0x9f] = {spc_illegal_op,},

		/* 0xa0 -> 0xaf */
		{spc_illegal_op,}, /* processed in the kernel module */
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{smc_move_medium,},
		{spc_illegal_op,},
		{spc_illegal_op,},

		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},

		/* 0xb0 -> 0xbf */
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},

		{smc_read_element_status,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},

		[0xc0 ... 0xdf] = {spc_illegal_op,},

		/* 0xe0 -> 0xef */
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{smc_initialize_element_status_with_range,},

		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},
		{spc_illegal_op,},

		[0xf0 ... 0xff] = {spc_illegal_op,},
	}
};
#pragma GCC diagnostic pop

__attribute__((constructor)) static void smc_init(void)
{
	device_type_register(&lunit, &smc_template);
}

/*
 * Update ops[xx] with new/updated/custom function 'f'
 */
void register_ops(struct lu_phy_attr *lu, int op,
			void *f, void *g, void *h)
{
	lu->scsi_ops->ops[op].cmd_perform = f;
	lu->scsi_ops->ops[op].pre_cmd_perform = g;
	lu->scsi_ops->ops[op].post_cmd_perform = h;
}

/*
 *
 * Process the SCSI command
 *
 * Called with:
 *	cdev          -> Char dev file handle,
 *	cdb           -> SCSI Command buffer pointer,
 *	struct mhvtl_ds -> general purpose data structure... Need better name
 *
 * Return:
 *	SAM status returned in struct mhvtl_ds.sam_stat
 */
static void processCommand(int cdev, uint8_t *cdb, struct mhvtl_ds *dbuf_p,
			useconds_t pollInterval)
{
	int err = 0;
	struct scsi_cmd _cmd;
	struct scsi_cmd *cmd;
	cmd = &_cmd;

	cmd->scb = cdb;
	cmd->scb_len = 16;	/* fixme */
	cmd->dbuf_p = dbuf_p;
	cmd->lu = &lunit;
	cmd->pollInterval = pollInterval;
	cmd->cdev = cdev;

	MHVTL_DBG_PRT_CDB(1, cmd);

	switch (cdb[0]) {
	case REPORT_LUNS:
	case REQUEST_SENSE:
	case MODE_SELECT:
	case INQUIRY:
		dbuf_p->sam_stat = SAM_STAT_GOOD;
		break;
	case MOVE_MEDIUM:
		/*
		 * This can change the vtl --show 
		 * since a cartridge can get loaded
		 */
		push_show = 1;
		break;
	default:
		if (cmd->lu->online == 0) {
			sam_not_ready(E_OFFLINE, &dbuf_p->sam_stat);
			return;
		}
		if (check_reset(&dbuf_p->sam_stat))
			return;
	}

	/* Skip main op code processing if pre-cmd returns non-zero */
	if (cmd->lu->scsi_ops->ops[cdb[0]].pre_cmd_perform)
		err = cmd->lu->scsi_ops->ops[cdb[0]].pre_cmd_perform(cmd, NULL);

	if (!err)
		dbuf_p->sam_stat = cmd->lu->scsi_ops->ops[cdb[0]].cmd_perform(cmd);

	/* Post op code processing regardless */
	if (cmd->lu->scsi_ops->ops[cdb[0]].post_cmd_perform)
		cmd->lu->scsi_ops->ops[cdb[0]].post_cmd_perform(cmd, NULL);

	return;
}

/*
 * Respond to messageQ 'list map' by sending a list of PCLs to messageQ
 */
static void list_map(struct q_msg *msg)
{
	struct list_head *slot_head = &smc_slots.slot_list;
	struct s_info *sp;
	char buf[MAXOBN];
	char *c = buf;
	*c = '\0';

	list_for_each_entry(sp, slot_head, siblings) {
		if (slotOccupied(sp) && sp->element_type == MAP_ELEMENT) {
			strncat(c, (char *)sp->media->barcode, MAX_BARCODE_LEN+1);
			MHVTL_DBG(2, "MAP slot %d full", sp->slot_location);
		} else {
			MHVTL_DBG(2, "MAP slot %d empty", sp->slot_location);
		}
	}
	MHVTL_DBG(2, "map contents: %s", buf);
	send_msg(buf, msg->snd_id);
}

/* Seems there is a new fucntion name "get_cart_type(...)"
 * return  1 for Data Cart, 2 for Cleaning Cart
 */

/* Check existing MAP & Storage slots for existing barcode */
int already_in_slot(char *barcode)
{
	struct list_head *slot_head = &smc_slots.slot_list;
	struct s_info *sp = NULL;
	int len;

	len = strlen(barcode);

	list_for_each_entry(sp, slot_head, siblings) {
		if (slotOccupied(sp)) {
			if (!strncmp((char *)sp->media->barcode, barcode, len)){
				MHVTL_DBG(3, "Match: [%s] [%s]",
					sp->media->barcode, barcode);
				return 1;
			} else {
				MHVTL_DBG(3, "No match: [%s] [%s]",
					sp->media->barcode, barcode);
			}
		}
	}
	return 0;
}

static struct s_info *locate_empty_map(void)
{
	struct s_info *sp = NULL;
	struct list_head *slot_head = &smc_slots.slot_list;

	list_for_each_entry(sp, slot_head, siblings) {
		if (!slotOccupied(sp) && sp->element_type == MAP_ELEMENT)
			return sp;
	}

	return NULL;
}

static struct m_info *lookup_barcode(struct lu_phy_attr *lu, char *barcode)
{
	struct list_head *media_list_head;
	struct m_info *m;
	int match;

	media_list_head = &((struct smc_priv *)lu->lu_private)->media_list;

	list_for_each_entry(m, media_list_head, siblings) {
		match = strncmp(m->barcode, barcode, MAX_BARCODE_LEN + 1);
		if (!match) {
			MHVTL_DBG(3, "Match barcodes: [%s] [%s]: %d",
				barcode, m->barcode, match);
			return m;
		}
	}

	return NULL;
}

static struct m_info *add_barcode(struct lu_phy_attr *lu, char *barcode, char *nsbarcode)
{
	struct list_head *media_list_head;
	struct m_info *m;

	if (strlen(barcode) > MAX_BARCODE_LEN) {
		MHVTL_ERR("Barcode \'%s\' exceeds max barcode lenght: %d",
				barcode, MAX_BARCODE_LEN);
		exit(1);
	}
	if (lookup_barcode(lu, barcode)) {
		MHVTL_ERR("Duplicate barcode %s.. Exiting", barcode);
		exit(1);
	}

	m = zalloc(sizeof(struct m_info));
	if (!m) {
		MHVTL_ERR("Out of memory allocating memory for barcode %s",
			  barcode);
		exit(-ENOMEM);
	}

	media_list_head = &((struct smc_priv *)lu->lu_private)->media_list;

	memset(m, 0, sizeof(struct m_info));

	snprintf((char *)m->barcode, MAX_BARCODE_LEN + 1, LEFT_JUST_16_STR,
					barcode);
	m->barcode[MAX_BARCODE_LEN] = '\0';

	/* Store original namespace barcode */
	if (nsbarcode) {
		snprintf((char *)m->nsbarcode, MAX_BARCODE_LEN + 1, LEFT_JUST_16_STR,
					nsbarcode);
		m->nsbarcode[MAX_BARCODE_LEN] = '\0';
	}
	
	m->cart_type = get_cart_type(barcode);
	if (!strncmp((char *)m->barcode, "NOBAR", 5))
		m->internal_status = INSTATUS_NO_BARCODE;
	else
		m->internal_status = 0;

	list_add_tail(&m->siblings, media_list_head);
	return m;
}


void push_bar(struct serials ** head_ref, char * serial)
{
	struct serials * new_node = (struct serials*)malloc(sizeof(struct serials));
	memset(new_node->name, 0, sizeof(new_node->name));
	strncpy(new_node->name, serial, 10);
	new_node->next = (*head_ref);
	(*head_ref) = new_node;
}

static struct s_info *get_empty_map_slot(void)
{
        struct s_info *sp = NULL;
        struct list_head *slot_head = &smc_slots.slot_list;

        list_for_each_entry(sp, slot_head, siblings) {
                if (sp->element_type == MAP_ELEMENT && !slotOccupied(sp))
                        return sp;
        }
        return NULL;
}


static int restore_cart_from_cloud_done()
{
  struct s_info *sp;
  struct m_info *mp = NULL;
  char *barcode = NULL;
  bool found = false;
  struct serial_list *tmp1, *tmp2, *tmp3;
  
  while(1) {
    if(list_empty(&import_serial_list.list))
      break;
    
    pthread_mutex_lock(&import_lock);
    list_for_each_entry(tmp1, &import_serial_list.list, list)
      if(tmp1->status == RES_CART_STATUS_CLDFLT_DONE)
	break;
    pthread_mutex_unlock(&import_lock);
    
    if(!tmp1 || tmp1->status != RES_CART_STATUS_CLDFLT_DONE)
      break;
    
    MHVTL_DBG(1, "Adding cart[%s] to map.", tmp1->name);
    sp = get_empty_map_slot();
    mp = NULL;
    if (sp) {
      if (sp->media) {
	if (!strncmp(sp->media->barcode, tmp1->name, strlen(tmp1->name))) {
	  continue;
	}
      }
      mp = lookup_barcode(&lunit, tmp1->name);
      if (!mp) {
	MHVTL_DBG(3,"add_barcode() [%s] to mp", tmp1->name);
	mp = add_barcode(&lunit, tmp1->name, tmp1->name);
      } else {
	MHVTL_DBG(1, "mp already present, mp->barcode is [%s]", mp->barcode);
      }
      
      snprintf((char *)mp->barcode, MAX_BARCODE_LEN, LEFT_JUST_16_STR, tmp1->name);
      mp->barcode[MAX_BARCODE_LEN] = '\0';
      mp->cart_type = MEDIA_TYPE_DATA;
      sp->status = STATUS_InEnab | STATUS_ExEnab | STATUS_Access | STATUS_ImpExp | STATUS_Full;
      setImpExpStatus(sp, OPERATOR);
      sp->media = mp;
      sp->media->internal_status = 0;
      found = true;
      
      pthread_mutex_lock(&import_lock);
      list_for_each_entry_safe(tmp2, tmp3, &import_serial_list.list, list)
	if(strcmp(tmp2->name, tmp1->name) == 0) {
	  list_del(&tmp2->list);
	  free(tmp2);
	  pthread_mutex_unlock(&import_lock);
	  break;
	}
      pthread_mutex_unlock(&import_lock);
      push_show = 1;
    } else {
      MHVTL_DBG(1,"get_empty_map_slot() did not return a slot.");
    }
  }
  
  if (found)
    map_activity();
  
  /* 
   * Set unit attention because there is a cart in the MAP slot and this
   * can now be moved to a regular slot and then to a tape drive for 
   * restore etc 
   */
  
  return 0;
}

static void *restore_cart_from_cloud(void *arg)
{
  /* 
   * This should be async -- it simply starts the CloudFault process and 
   * once done restore_cart_from_cloud_done() is called. How you do it 
   * is up to your imagination! :-)
   */
  char cart_files[3][5] = { "data", "meta", "indx" };
  char *barcode = (char *)arg;
  int i, rc = 0;
  uint8_t		*tmp_buf = NULL;
  char export_file[256];
  struct serial_list *tmp_serial, *tmp;
  struct stat 	stats;
  
  sprintf(export_file, "%s/%s/export", cart_home, barcode);
  
  tmp_buf = (uint8_t *)malloc(sizeof(uint8_t) * MAX_IOSZ);
  memset(tmp_buf, 0, MAX_IOSZ);
  
  pthread_mutex_lock(&import_lock);
  list_for_each_entry(tmp, &import_serial_list.list, list){
    if(strcmp(tmp->name, barcode) == 0){
      rc = 1;
      pthread_mutex_unlock(&import_lock);
      goto failed;
    }
  }
  
  tmp_serial = (struct serial_list *)malloc(sizeof(struct serial_list));
  sprintf(tmp_serial->name, barcode);
  tmp_serial->status = RES_CART_STATUS_CLDFLT_STARTED;
  list_add_tail(&(tmp_serial->list), &(import_serial_list.list));
  pthread_mutex_unlock(&import_lock);
  
  if (stat(export_file, &stats) < 0 && errno == ENOENT) {
    MHVTL_LOG("%s not found, cart %s may never got exported", export_file, barcode);
    rc = 1;
    goto failed;
  }
  
  pthread_mutex_lock(&import_lock);
  tmp_serial->status = RES_CART_STATUS_CLDFLT_DONE;	
  unlink(export_file);
  pthread_mutex_unlock(&import_lock);
  
 failed:
  if(tmp_buf)
    free(tmp_buf);

  return rc;
}

static int load_map(struct q_msg *msg)
{
  char *barcode;
  int i, ret = 0;
  int str_len;
  char *text = &msg->text[9];	/* skip past "load map " */
  char export_marker[256] = "";
  pthread_t	restore_thread;
  struct serial_list *tmp = NULL;
  struct stat 	stats;
  
  str_len = strlen(text);
  barcode = NULL;
  for (i = 0; i < str_len; i++) {
    if (isalnum(text[i])) {
      barcode = &text[i];
      break;
    }
  }
  
  if (!barcode) {
    send_msg("Bad barcode", msg->snd_id);
    return 0;
  }
  
  truncate_spaces(barcode, MAX_BARCODE_LEN + 1);
  
  if (already_in_slot(barcode)) {
    send_msg("Barcode already in library", msg->snd_id);
    return 0;
  }
  
  if (strlen(barcode) > MAX_BARCODE_LEN) {
    send_msg("Barcode length too long", msg->snd_id);
    return 0;
  }
  
  pthread_mutex_lock(&import_lock);

  if(!list_empty(&import_serial_list.list)){
    list_for_each_entry(tmp, &import_serial_list.list, list){
      if(strcmp(tmp->name, barcode) == 0){
	send_msg("Cartidge already in the process of being imported from"
		 " cloud...", msg->snd_id);
	pthread_mutex_unlock(&import_lock);
	return 0;
      }
    }
  }
  
  sprintf(export_marker, "%s/pool/%s", home_directory, barcode);
  if (stat(export_marker, &stats) < 0 && errno == ENOENT) {
    send_msg("barcode not in library nor exported to cloud, cannot import.",
	     msg->snd_id);
    pthread_mutex_unlock(&import_lock);
    return 0;
  } else {
    /* TODO Check if all three files of the cartridge are completely in cloud
     *  if not, skip the import
     *  Make a http call (dir_status) to policy_mgr to know if all files 
     *  (data, meta, indx) are in cloud
     */
  }
  
  pthread_mutex_unlock(&import_lock);
  
  /*
   * This is where you pull the file the cloud. Meaning you need to 
   * cloudfault the file in pool directory. Once it is restored in 
   * the pool directory, move it into the carts directory and place 
   * it in the map slot.
   *
   */
  ret = pthread_create(&restore_thread, NULL, restore_cart_from_cloud,
		       (void *)barcode);
  if (ret) {
    send_msg("Restoring of cart may be already started ...", msg->snd_id);
    return 0;
  }
  
  ret = pthread_detach(restore_thread);
  if(ret)
    MHVTL_LOG("Detach of restore thread failed for %s", barcode);
  
  send_msg("Starting restore of cartridge from cloud ...", msg->snd_id);
  
  return 0;
}

/*
 * Dump all details of tape drives and carts for fast `vtl --show`. 
 * [ See bug 23788 for why I did this - why I don't want mtx and 
 *   sginfo to interfere with regular i/o ]. Hmm! since most of this
 * stuff is static, let me write to the file only when a load or 
 * unload event occurs or when new carts are added.
 */

static void vtl_show(struct lu_phy_attr *lu)
{
  struct list_head *drive_list_head;
  struct d_info *d;
  struct list_head *slot_head = &smc_slots.slot_list;
  struct s_info *sp = NULL;
  char *nl = "Not loaded";
  int nc = 0, n = 0, stop = 0;
  DIR *pool_d;
  char pool[256], export_file[256];
  struct stat 	stats;
  char buff[4096] = {0};
  int len = 0;
  
  if (!push_show)
    return;
  
  push_show = 0;
  
  MHVTL_LOG("Pushing vtl --show ...");
  
  /*
   * Well-known location
   */
  sprintf(g_f, "/tmp/info.%ld", my_id);
  g_fp = fopen(g_f, "w");
  if (!g_fp)
    return;
  
  drive_list_head = &((struct smc_priv *)lu->lu_private)->drive_list;
  
  /*
   * Count number of carts
   */
  list_for_each_entry(sp, slot_head, siblings) {
    if ((sp->element_type & (STORAGE_ELEMENT|MAP_ELEMENT|DATA_TRANSFER)) &&
	slotOccupied(sp))
      nc++;
  }
	
  fprintf(g_fp, " medi\t%.5s\t%.11s\t%s\t%5d %dGB\t%ld%s\n",
	  lu->vendor_id,
	  lu->product_id, 
	  lu->lu_serial_no, 
	  nc, tape_size/1024,
	  my_id,
	  g_repl_salt[0] ? " [R]" : "");

  /*
   * Show all tape drives
   */
  list_for_each_entry(d, drive_list_head, siblings) {
    fprintf(g_fp, " tape\t%s\t%s\t%s\t%.10s\t%ld\n",
	    d->inq_vendor_id,
	    d->inq_product_id,
	    d->inq_product_sno,
	    d->slot ?
	    (d->slot->media ? 
	     d->slot->media->barcode :
	     nl)
	    : nl,
	    d->drv_id);
  }

  /*
   * Show all cartridges
   */
  n=0;
  fprintf(g_fp, "\nCartridges available in Storage Slots:\n ");
  list_for_each_entry(sp, slot_head, siblings) {
    if (sp->element_type == STORAGE_ELEMENT) {
      if (stop)
	continue;
      fprintf(g_fp, "%.9s %s", 
	      slotOccupied(sp) ? sp->media->barcode : "[ empty ]",
	      ((++n % 8) == 0) ? "\n " : "");
      
      /* 
       * Show up to 80 cartridges and not create
       * a mess by showing all of the cartridges
       */
      if (n > 80) {
	fprintf(g_fp, " ... and more!");
	stop++;
      }
    }
  }
  
  fprintf(g_fp, "\nCartridges available in MAP Slot:\n ");
  n=0;
  list_for_each_entry(sp, slot_head, siblings) {
    if (sp->element_type == MAP_ELEMENT) {
      fprintf(g_fp, "%.9s %s", 
	      slotOccupied(sp) ? sp->media->barcode : "[ empty ]",
	      ((++n % 8) == 0) ? "\n " : "");
    }
  }
  
  fprintf(g_fp, "\nTotal: %d cartridges available in Library.\n", nc);
  n=0;
  fprintf(g_fp, "\nCartridges exported to the %s:\n ",
	  (archive) ? "Glacier" : "Cloud");

  sprintf(pool, "%s/pool", home_directory);
  pool_d = opendir(pool);
  if (!pool_d) {
    MHVTL_ERR("Can not open dir %s/pool: %s", home_directory, strerror(errno));
    goto done;
  } else {
    while(1) {
      struct dirent * entry;
      entry = readdir(pool_d);
      if (!entry) {
	closedir(pool_d);
	MHVTL_DBG(1, "Done all pool dir ");
	break;
      } else if (strcmp(entry->d_name,".") == 0 ||
		 strcmp(entry->d_name,"..") == 0 ) {
	continue;
      }
      
      sprintf(export_file, "%s/%s/export", cart_home, entry->d_name);
      if (!stat(export_file, &stats)) {
	fprintf(g_fp, "%.9s %s", entry->d_name,
		((++n % 8) == 0) ? "\n " : "");
      }
      
      if (archive) {
	char export_marker_file[256];
	char *xattr_restore_job_status = "user.restore.job.status";
	char *xattr_restore_data_expiry = "user.restore.data.expiry.timestamp";
	char xattr_val2[128] = {0}, xattr_val3[128] = {0};
	ssize_t xattr_val2_len = 0, xattr_val3_len = 0;
	
	sprintf(export_marker_file, "%s/%s", pool, entry->d_name);
	
	xattr_val2_len = getxattr(export_marker_file, xattr_restore_job_status,
				  &xattr_val2[0], sizeof(xattr_val2));
	if (xattr_val2_len > 0) {
	  xattr_val2[xattr_val2_len] = '\0';
	  xattr_val3_len = getxattr(export_marker_file, xattr_restore_data_expiry,
				    &xattr_val3[0], sizeof(xattr_val3));
	  if (xattr_val3_len > 0) {
	    // Max characters occupied per cart's restore status in the output buffer
	    //
	    // 03 - Ctrl characters 
	    // 09 - Cart Name length
	    // 40 - Restore status string's max length
	    // 30 - RFC1123 format timestamp max length
	    //
	    if (sizeof(buff) - len > (3+9+40+30)) {
	      if (strcasestr(xattr_val2, "OPERATION_SUCCESSFUL")) {
		len += snprintf(buff + len, sizeof(buff) - len, "%s\t%s\t%s\n",
				entry->d_name, xattr_val2, xattr_val3);
	      } else {
		len += snprintf(buff + len, sizeof(buff) - len, "%s\t%s\t%s\n",
				entry->d_name, xattr_val2, "-N/A-");
	      }
	    } else { // if (sizeof(buff) - len > (3+9+40+30)) { ... } else {
	      if (strcasestr(xattr_val2, "OPERATION_SUCCESSFUL")) {
		MHVTL_ERR("VTL Show ouput truncated: %s\t%s\t%s", entry->d_name,
			  xattr_val2, xattr_val3);
	      } else {
		MHVTL_ERR("VTL Show ouput truncated: %s\t%s\t%s", entry->d_name,
			  xattr_val2, "-N/A-");
	      }
	    } // if (sizeof(buff) - len > (3+9+40+30)) { ... } else { ... }
	  } // if (xattr_val3_len > 0) { ... }
	  else {
	    MHVTL_DBG(3, "Retrieval of value of the extended attribute '%s' "
		      "associated with the path '%s' has failed: %s",
		      xattr_restore_data_expiry, export_marker_file,
		      strerror(errno));
	  }
	} // if (xattr_val2_len > 0) { ... }
	else {
	  MHVTL_DBG(3, "Retrieval of value of the extended attribute '%s' "
		    "associated with the path '%s' has failed: %s",
		    xattr_restore_job_status, export_marker_file,
		    strerror(errno));
	}
      } // if (archive) { ... }
    }
	  
    fprintf(g_fp, "\nTotal: %d cartridge(s) available in %s.\n", n,
	    (archive) ? "Glacier" : "Cloud");
  }

  if (archive) {
    char header[128] = {0};

    snprintf(header, sizeof(header), "%s\t%s\t%s\n",
	     "[Cart    ]",
	     "[Restore Operation Status          ]",
	     "[Restored Data Expiry timestamp    ]");

    fprintf(g_fp, "\nCart(s) Archive Restore Status:\n");
    fprintf(g_fp, "%s\n", header);
    fprintf(g_fp, "%s\n", buff);
  }
  
 done:
  fclose(g_fp);
  g_fp = NULL;
}

static struct s_info *previous_storage_slot(struct s_info *s,
			struct list_head *slot_head);

/* 
 * Reinstatement case - vtltape has died and has been
 * restarted. It does *not* know that it had a cartridge.
 * But the vtllibrary knows. So, in this routine we just 
 * "unload" the cart back to its original position. 
 *
 * vtllibrary reinstatement -- just send a message to 
 * all tape drives, "unload_ua"
 * 
 * Thought - maybe I can pass a "lload" to the tape drive
 * and load the cartridge in the first case. TODO. TTxxx.
 */ 
static void unload_td(struct lu_phy_attr *lu, int tdid)
{
  struct smc_priv *lu_priv;
  struct list_head *slot_head;
  struct list_head *drive_head;
  struct s_info *sp;	/* Slot Pointer */
  struct d_info *dp;	/* Drive Pointer */

  lu_priv = lu->lu_private;
  drive_head = &lu_priv->drive_list;
  slot_head = &lu_priv->slot_list;
  
  /* 
   * Unload the cartridge in a tape drive to
   * its previous location. 
   */
  list_for_each_entry(dp, drive_head, siblings) {
    /* vtllibrary was restarted */	
    if (tdid == -1) {
      /* Send 'unload' message and set unit attention */
      MHVTL_LOG("Unloading drive %ld!", dp->drv_id);
      send_msg("unload_ua", dp->drv_id);
      /* a vtltape was restarted */
    } else if (slotOccupied(dp->slot) && (int) dp->drv_id == tdid) {
      sp = dp->slot;
      MHVTL_LOG("Found %s in drive %d from slot %d -- unloading!",
		sp->media->barcode,
		sp->slot_location - START_DRIVE + 1,
		sp->last_location);

      /* if (tdreq_unload_tape) { */
      /* 	MHVTL_LOG("Unloading drive %ld! on tape unload req from tape drive", */
      /* 		  dp->drv_id); */
      /* 	send_msg("unload", dp->drv_id); */
      /* 	tdreq_unload_tape = false; */
      /* } */

      /* 
       * Return the slot to its original 
       * position from where it was loaded into
       * the tape drive, td.
       */
      unload_drive_on_shutdown(sp, previous_storage_slot(sp, slot_head)); 
    }
  }
  
  return;
}

/* add new slot && assignment && initialization memory */
static void add_storage_slot(struct q_msg *msg)
{
	int buffer_size;
	int slt_no;
	char message[20];
	struct s_info *sp1 = NULL;
	struct smc_priv *smc_p = lunit.lu_private;

	sp1 = add_new_slot(&lunit);

	smc_p->num_storage++;
	slt_no = smc_p->num_storage;
	sp1->element_type = STORAGE_ELEMENT;
	sp1->status = STATUS_Access;
	sp1->slot_location = slt_no + smc_p->pm->start_storage - 1;

	/* Slot status to Empty */
	setSlotEmpty(sp1);

	init_smc_log_pages(&lunit);
	init_smc_mode_pages(&lunit);
	MHVTL_DBG(1, "add slot && init smc");

	/* malloc a big enough buffer to fit worst case read element status */
	buffer_size = (smc_slots.num_drives + smc_slots.num_picker
			+ smc_slots.num_map + smc_slots.num_storage) * 80;
	buffer_size = max(SMC_BUF_SIZE, buffer_size);
	smc_slots.bufsize = buffer_size;
	MHVTL_DBG(1, "Setting buffer size to %d", buffer_size);

	sprintf(message, "slot=%d", slt_no);
	send_msg(message, msg->snd_id);
	return;
}

/*
 * Respond to messageQ 'empty map' by clearing 'ocuplied' status in map slots.
 * Return 0 on failure, non-zero - success.
 */
/* static int empty_map(struct q_msg *msg) */
/* { */
/* 	struct s_info *sp; */
/* 	struct list_head *slot_head = &smc_slots.slot_list; */
/* 	if (smc_slots.cap_closed) { */
/* 		MHVTL_DBG(1, "MAP slot empty failed - CAP Not open"); */
/* 		send_msg("Can't empty map while MAP is closed", msg->snd_id); */
/* 		return 0; */
/* 	} */
/* 	list_for_each_entry(sp, slot_head, siblings) { */
/* 		if (slotOccupied(sp) && sp->element_type == MAP_ELEMENT) { */
/* 			setSlotEmpty(sp); */
/* 			MHVTL_DBG(2, "MAP slot %d emptied", */
/* 					sp->slot_location - */
/* 						smc_slots.pm->start_map); */
/* 		} */
/* 	} */
/* 	send_msg("OK", msg->snd_id); */
/* 	return 1; */
/* } */

/* Extract the id of the tape sending notification a tape was ejected
 * Set the 'access' bit in the READ_ELEMENT_STATUS page
 */
/* static int set_access_bit(struct q_msg *msg) */
/* { */
/*   struct s_info *sp; */
/*   struct list_head *slot_head = &smc_slots.slot_list; */
  
/*   list_for_each_entry(sp, slot_head, siblings) { */
/*     if (slotOccupied(sp) && sp->element_type == DATA_TRANSFER) { */
/*       if (sp->drive->drv_id == msg->snd_id) { */
/* 	setAccessStatus(sp, 1); */
/* 	MHVTL_DBG(2, "Enabling access bit for drive id %ld", sp->drive->drv_id); */
/*       } */
/*     } */
/*   } */
/*   return 0; */
/* } */

/*
 * Return 1, exit program
 */
static int processMessageQ(struct q_msg *msg, struct mhvtl_ctl *ctl, int cdev)
{
  MHVTL_DBG(1, "Sender id: %ld, msg : %s", msg->snd_id, msg->text);
  
  if (!strncmp(msg->text, "crashpoint", 10)) {
    uint64_t p = 0;
    
    /* Set the crash point and go */
    sscanf(msg->text, "crashpoint %" PRIx64, &p);
    if (p >= 1 && p < 64) {
      MHVTL_LOG("Added crash point at # 0x%lx\n", p);
      lunit.cp |= 1 << (p - 1);
    }
  }
  
  if (!strncmp(msg->text, "silent", 6)) {
    verbose=0;
    cdbonly=0;
  }
  
  if (!strncmp(msg->text, "verbose", 7)) {
    if (verbose)
      verbose--;
    else
      verbose = 3;
    
    MHVTL_LOG("verbose: %s at level %d", verbose ? "enabled" : "disabled",
	      verbose);
  }

  if (!strncmp(msg->text, "cdbonly", 7)) {
    cdbonly = 1;
  }

  if (!strncmp(msg->text, "create_carts", 12)) {
    ioctl(cdev, VTL_BUSY_LU, ctl);
    create_carts(msg);
    ioctl(cdev, VTL_UNBUSY_LU, ctl);
  }
  
  if (!strncmp(msg->text, "vtl_show", 8)) {
    push_show = 1;
    vtl_show(&lunit);
  }

  if (!strncmp(msg->text, msg_eject, strlen(msg_eject)))  {
    MHVTL_LOG("Got eject message from tape drive: %d", msg->snd_id);
    /* int tdid = msg->snd_id; */
    /* if (strncasecmp(lunit.vendor_id, "quest", strlen("quest")) == 0) { */
    /*   MHVTL_LOG("Move medium from drive to storage slot for VTL vendor: %s", */
    /* 		lunit.vendor_id); */
    /*   tdreq_unload_tape = true; */
    /*   unload_td(&lunit, tdid); */
    /*   push_show = 1;    /\* poke it *\/ */
    /* } else { */
    /*   send_msg("unload", tdid); */
    /* } */
  }
  
  if (!strncmp(msg->text, "unload_td", 9)) {
    int tdid;
    
    if (sscanf(msg->text, "unload_td %d", &tdid) != 1) {
      MHVTL_ERR("unload_td: wrong arguments in message: %s\n", msg->text);
    } else {
      unload_td(&lunit, tdid);
      push_show = 1;	/* poke it */
    }
  }
  
  /* A gentleman's request to quit graciously */
  if (!strncmp(msg->text, "exit", 4)) {
    return 1;
  }
  
  /* 
   * Load a cartridge into a MAP slot manually -- meaning in our case we
   * get a file from the cloud which represents a cartridge and then
   * add it to an avaiable MAP slot
   */
  if (!strncmp(msg->text, "load map ", 9)) {
    load_map(msg);
  }
  
  return 0;
}

static struct d_info *lookup_drive(struct lu_phy_attr *lu, unsigned int drive_no)
{
	struct list_head *drive_list_head;
	struct d_info *d;
	/* uint32_t slot_offset; */

	drive_list_head = &((struct smc_priv *)lu->lu_private)->drive_list;
	/* slot_offset = ((struct smc_priv *)lu->lu_private)->pm->start_drive; */

	/* /\* Drive numbering starts from 1, decrement slot_offset *\/ */
	/* slot_offset--; */

	list_for_each_entry(d, drive_list_head, siblings) {
		/* MHVTL_DBG(3, "Slot location: %d, offset + drive_no: %d", */
		/* 		d->slot->slot_location, slot_offset + drive_no); */
		MHVTL_DBG(3, "Slot location: %d, drive_no: %d", d->slot->slot_location,
			  drive_no);
		/* if (d->slot->slot_location == slot_offset + drive_no) */
		if (d->slot->slot_location == drive_no)
		    return d;
	}

	return NULL;
}

struct s_info *add_new_slot(struct lu_phy_attr *lu)
{
	struct s_info *new;
	struct list_head *slot_list_head;

	slot_list_head = &((struct smc_priv *)lu->lu_private)->slot_list;

	new = zalloc(sizeof(struct s_info));
	if (!new) {
		MHVTL_ERR("Could not allocate memory for new slot struct");
		exit(-ENOMEM);
	}

	list_add_tail(&new->siblings, slot_list_head);
	return new;
}

/* Open device config file and update device information
 */
static void update_drive_details(struct lu_phy_attr *lu)
{
	unsigned int slot;
	struct d_info *dp;
	struct s_info *sp;
	struct smc_priv *smc_p = lu->lu_private;
	char buf[32];

	for (slot = 1; slot <= no_of_drives; slot++) {

		dp = lookup_drive(lu, slot);

		if (!dp) {
			dp = zalloc(sizeof(struct d_info));
			if (!dp) {
				MHVTL_ERR("Couldn't malloc memory (dp) ...");
				exit(-ENOMEM);
			}
			sp = add_new_slot(lu);
			sp->element_type = DATA_TRANSFER;
			dp->slot = sp;
			sp->drive = dp;
			list_add_tail(&dp->siblings,
					&smc_p->drive_list);
		}

		dp->drv_id = (my_id*60)-40 + slot;
		sprintf(buf, "%s_%.2d", g_serial_base, slot);
		strncpy(dp->inq_product_sno, replsalt(buf), 10);
		rmnl(dp->inq_product_sno, ' ', 10);
		// Moving to TD8 from TD4
		strncpy(dp->inq_product_id, "ULT3580-TD8", 11);
		rmnl(dp->inq_product_id, ' ', 16);

		strncpy(dp->inq_vendor_id, "IBM", 3);
		rmnl(dp->inq_vendor_id, ' ', 8);
	}

	return;
}

/*
 * Convert base10 to base36
 */
static char *__get_base36(int base10, char *alphanum)
{
        char *base36char = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
        int i;
        for(i=3-1; i>=0; --i) {
                alphanum[i] = base36char[base10 % 36];
                base10 /= 36;
        }
        return alphanum;
}

/* 
 * This function returns base10 of the alphanum
 * returns 0 when aplhanum has invalid number of characters
 */
static int __get_base10(char *alphanum)
{
        char *base36char = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
        int i, j, k, len;
        double base10 = 0, power;
        len = strlen(alphanum);
	if (len < 1 || len > 3)
		return 0;
        for (i=len-1, k=0; i>=0; i--,k++) {
                for(j=0; j<36; j++){
                        if(alphanum[i] == base36char[j]) {
                                power = pow(36.0, k);
                                base10 = base10 + power * j;
                                break;
                        }
                }
        }
        return (int)base10;
}


void mergesort(struct serials **headref)
{
	struct serials *head = *headref;
	struct serials *a;
	struct serials *b;

	if ((head == NULL) || (head->next == NULL))
		return;
	split(head, &a, &b);

	mergesort(&a);
	mergesort(&b);
	*headref = sortedmerge(a, b);
}

struct serials *sortedmerge(struct serials *a, struct serials *b)
{
	struct serials *result = NULL;
	if (a == NULL)
		return b;
	else if (b == NULL)
		return a;

	if (strcmp(a->name, b->name) <= 0)
	{
		result = a;
		result->next = sortedmerge(a->next, b);
	} else
	{
		result = b;
		result->next = sortedmerge(a, b->next);
	}
	return result;
}

void split(struct serials *source, struct serials **frontref,
                struct serials **backref)
{
	struct serials *fast;
	struct serials *slow;
	if (source == NULL || source->next == NULL)
	{
		*frontref = source;
		*backref = NULL;
	} else
	{
		slow = source;
		fast = source->next;
		while (fast != NULL)
		{
		  fast = fast->next;
		  if (fast != NULL)
		  {
		    slow = slow->next;
		    fast = fast->next;
                  }
		}
		*frontref = source;
		*backref = slow->next;
		slow->next = NULL;
	}
}

void free_serials(struct serials *list)
{
	struct serials *tmp;
	while (list != NULL)
	{
		tmp = list;
		list = list->next;
		free(tmp);
		tmp = NULL;
	}
}

/*
 * Return 0 - no address space conflict
 * Return 1 - overlap address with another slot type
 */
static int check_overflow(struct lu_phy_attr *lu, int slot, char type)
{
	struct smc_priv *smc_p;
	int co;

	smc_p = lu->lu_private;
	co = 0;

	switch (type) {
	case MAP_ELEMENT:
		co = slot + smc_p->pm->start_map;
		if (smc_p->pm->start_map < smc_p->pm->start_storage &&
					co > smc_p->pm->start_storage) {
			MHVTL_LOG("MAP: %d, overlaps with storage slot", slot);
			return 1;
		}
		if (smc_p->pm->start_map < smc_p->pm->start_picker &&
					co > smc_p->pm->start_picker) {
			MHVTL_LOG("MAP: %d, overlaps with Picker slot", slot);
			return 1;
		}
		if (smc_p->pm->start_map < smc_p->pm->start_drive &&
					co > smc_p->pm->start_drive) {
			MHVTL_LOG("MAP: %d, overlaps with Drives", slot);
			return 1;
		}
		break;
	case DATA_TRANSFER:
		co = slot + smc_p->pm->start_drive;
		if (smc_p->pm->start_drive < smc_p->pm->start_storage &&
					co > smc_p->pm->start_storage) {
			MHVTL_LOG("Drive: %d, overlaps with storage slot",
					slot);
			return 1;
		}
		if (smc_p->pm->start_drive < smc_p->pm->start_picker &&
					co > smc_p->pm->start_picker) {
			MHVTL_LOG("Drive: %d, overlaps with picker slot",
					slot);
			return 1;
		}
		if (smc_p->pm->start_drive < smc_p->pm->start_map &&
					co > smc_p->pm->start_map) {
			MHVTL_LOG("Drive: %d, overlaps with MAP slot",
					slot);
			return 1;
		}
		break;
	case MEDIUM_TRANSPORT:
		co = slot + smc_p->pm->start_picker;
		if (smc_p->pm->start_picker < smc_p->pm->start_map &&
					co > smc_p->pm->start_map) {
			MHVTL_LOG("Picker slot: %d overlaps with MAP", slot);
			return 1;
		}
		if (smc_p->pm->start_picker < smc_p->pm->start_drive &&
					co > smc_p->pm->start_drive) {
			MHVTL_LOG("Picker slot: %d overlaps with drives", slot);
			return 1;
		}
		if (smc_p->pm->start_picker < smc_p->pm->start_storage &&
					co > smc_p->pm->start_storage) {
			MHVTL_LOG("Picker slot: %d overlaps with Storage",
					slot);
			return 1;
		}
		break;
	case STORAGE_ELEMENT:
		co = slot + smc_p->pm->start_storage;
		if (smc_p->pm->start_storage < smc_p->pm->start_map &&
					co > smc_p->pm->start_map) {
			MHVTL_LOG("Storage slot: %d, overlaps with MAP", slot);
			return 1;
		}
		if (smc_p->pm->start_storage < smc_p->pm->start_picker &&
					co > smc_p->pm->start_picker) {
			MHVTL_LOG("Storage slot: %d, overlaps with picker",
					slot);
			return 1;
		}
		if (smc_p->pm->start_storage < smc_p->pm->start_drive &&
					co > smc_p->pm->start_drive) {
			MHVTL_LOG("Storage slot: %d, overlaps with drives",
					slot);
			return 1;
		}
		break;
	}
	return 0;
}

void init_drive_slot(struct lu_phy_attr *lu, int slt, char *s)
{
	struct s_info *sp = NULL;
	struct d_info *dp = NULL;
	struct smc_priv *smc_p = lu->lu_private;

	if (check_overflow(lu, slt, DATA_TRANSFER))
		return;

	dp = lookup_drive(lu, slt);
	if (!dp) {
		dp = zalloc(sizeof(struct d_info));
		if (!dp) {
			MHVTL_ERR("init_drive_slot: Couldn't malloc memory");
			exit(-ENOMEM);
		}
		sp = add_new_slot(lu);
		sp->element_type = DATA_TRANSFER;
		dp->slot = sp;
		sp->drive = dp;
		list_add_tail(&dp->siblings, &smc_p->drive_list);
	}
	dp->slot->slot_location = slt + smc_p->pm->start_drive - 1;
	dp->slot->status = STATUS_Access;
	smc_p->num_drives++;
	if (strlen(s)) {
		strncpy(dp->inq_product_sno, s, 10);
		MHVTL_DBG(2, "Drive s/no: %s", s);
	}
	MHVTL_DBG(3, "Slot: %d, start_drive: %d, slot_location: %d",
			slt, smc_p->pm->start_drive, dp->slot->slot_location);
}

void init_map_slot(struct lu_phy_attr *lu, int slt, char *barcode)
{
	struct s_info *sp = NULL;
	struct smc_priv *smc_p = lu->lu_private;

	if (check_overflow(lu, slt, MAP_ELEMENT))
		return;

	sp = add_new_slot(lu);
	sp->element_type = MAP_ELEMENT;
	smc_p->num_map++;

	sp->slot_location = slt + smc_p->pm->start_map - 1;
	sp->status = STATUS_InEnab | STATUS_ExEnab |
				STATUS_Access | STATUS_ImpExp;

	if (strlen(barcode)) {
		MHVTL_DBG(2, "Barcode %s in MAP %d", barcode, slt);
		sp->media = add_barcode(lu, barcode, NULL);
		sp->status |= STATUS_Full;
	}
}

void init_transport_slot(struct lu_phy_attr *lu, int slt, char *barcode)
{
	struct s_info *sp = NULL;
	struct smc_priv *smc_p = lu->lu_private;

	if (check_overflow(lu, slt, MEDIUM_TRANSPORT))
		return;

	sp = add_new_slot(lu);
	sp->element_type = MEDIUM_TRANSPORT;
	smc_p->num_picker++;
	sp->slot_location = slt + smc_p->pm->start_picker - 1;
	sp->status = 0;

	if (strlen(barcode)) {
		MHVTL_DBG(2, "Barcode %s in Picker %d", barcode, slt);
		sp->media = add_barcode(lu, barcode, NULL);
		sp->slot_location = slt + smc_p->pm->start_picker - 1;
		sp->status |= STATUS_Full;
	}
}

void init_storage_slot(struct lu_phy_attr *lu, int slt, char *barcode, char *nsbarcode)
{
	struct s_info *sp = NULL;
	struct smc_priv *smc_p = lu->lu_private;

	if (check_overflow(lu, slt, STORAGE_ELEMENT))
		return;

	sp = add_new_slot(lu);
	sp->element_type = STORAGE_ELEMENT;
	smc_p->num_storage++;
	sp->status = STATUS_Access;
	sp->slot_location = slt + smc_p->pm->start_storage - 1;
	if (strlen(barcode) && strcmp(barcode, "") != 0) {
		MHVTL_DBG(1, "Barcode %s in slot %d", barcode, slt);
		sp->media = add_barcode(lu, barcode, nsbarcode);
		/* Slot full */
		sp->status |= STATUS_Full;
	} else {
		MHVTL_LOG("initialize empty slot %d", slt);
		setSlotEmpty(sp);
		sp->media = NULL;
		sp->last_location = 0;
	}
}

static void __init_slot_info(struct lu_phy_attr *lu, int type)
{
  unsigned int slot, map;
  
  switch(type) {
  case DATA_TRANSFER:
    for (slot = 1; slot <= no_of_drives; slot++) {
      init_drive_slot(lu, slot, "");
    }
    break;
    
  case MAP_ELEMENT:
    /* 
     * Create 32 empty MAP (media access port) 
     * OR CAP (cartridge access slots). We can move
     * a cartridge from the regular slot to a MAP 
     * slot in order to export it (to the cloud etc). 
     * When importing, a cartridge will appear in the 
     * MAP slot. We then move the cartridge from the MAP
     * slot to a regular slot. Command to use for manually
     * doing this: `mtx -f <loader_dev> trasnfer 
     * <storage-element-number> <storage-element-number> 
     */
    for (map = 1; map <= 32; map++) {
      init_map_slot(lu, map, "");
    }
    break;
    
  case MEDIUM_TRANSPORT:
    /* No picker elements */
    init_transport_slot(lu, 1, "");
    break;
    
  case STORAGE_ELEMENT:
    MHVTL_DBG(1, "init Storage element");
    DIR * cart_d;
    struct serials *barcodes, *bp;
    barcodes = NULL;
    
    cart_d = opendir(cart_home);
    if (!cart_d) {
      MHVTL_ERR("Can not open dir %s", cart_home);
      return;
    }
    
    /* initialize gslots to 0 */
    while(1) {
      struct dirent * entry;
      struct stat st;
      int export_fd = -1;
      char export_marker[1024];
      
      entry = readdir(cart_d);
      if (!entry) {
	closedir(cart_d);
	MHVTL_DBG(1, "Done all cart dir ");
	break;
	/* 
	 * NOTE: Underlying filesystem is returning 
	 * d_type NULL - assume all are directories
	 * } else if (entry->d_type != DT_DIR || 
	 */
      } else if (strcmp(entry->d_name,".") == 0
		 || strcmp(entry->d_name,"..") == 0 ) {
	continue;
      }

      snprintf(export_marker, ARRAY_SIZE(export_marker), "%s/%s/export",
	       cart_home, entry->d_name);
      export_fd = open(export_marker, O_RDWR|O_LARGEFILE);
      if (export_fd != -1) {
	MHVTL_LOG("Found export marker skipping this cartridge [%s].", entry->d_name);
	close(export_fd);
	push_bar(&barcodes, "");
      } else {
	push_bar(&barcodes, entry->d_name);
      }
    } // while(1) { ... }
    
    mergesort(&barcodes);
    
    gslots = 0;
    for (bp = barcodes; bp != NULL; bp = bp->next) {
      gslots++;
      if (strcmp(bp->name, "")==0) {
	init_storage_slot(lu, gslots, "", "");
      } else {
	init_storage_slot(lu, gslots, replsalt(bp->name), bp->name);
      }
      
      MHVTL_DBG(1, "init Storage element:%d", gslots); 
    }
    
    if (barcodes)
      free_serials(barcodes);
    
    break; // case STORAGE_ELEMENT:
    
  default:
    MHVTL_ERR("%s: Unknown type :%d", __func__, type);
    break;
    
  } // switch(type) { ... }

  return;
}

/* Linked list data needs to be built in slot order */
void init_slot_info(struct lu_phy_attr *lu)
{
	int i;
	struct smc_type_slot arr[4];

	sort_library_slot_type(lu, &arr[0]);

	for (i = 0; i < 4; i++)
		__init_slot_info(lu, arr[i].type);
}

#ifndef STK

static void update_vpd_83(struct lu_phy_attr *lu, void *p)
{
  struct vpd *vpd_pg = lu->lu_vpd[PCODE_OFFSET(0x83)];
  uint8_t *d;
  int num;
  char *ptr;
  int len, j;

  d = vpd_pg->data;
  
  d[0] = 2;
  d[1] = 1;
  d[2] = 0;
  num = VENDOR_ID_LEN + PRODUCT_ID_LEN + 10;
  d[3] = num;
  
  memcpy(&d[4], &lu->vendor_id, VENDOR_ID_LEN);
  memcpy(&d[12], &lu->product_id, PRODUCT_ID_LEN);
  memcpy(&d[28], &lu->lu_serial_no, 10);
  len = (int)strlen(lu->lu_serial_no);
  ptr = &lu->lu_serial_no[len];

  num += 4;
  /* NAA IEEE registered identifier (faked) */
  d[num] = 0x1;	/* Binary */
  d[num + 1] = 0x3;
  d[num + 2] = 0x0;
  d[num + 3] = 0x8;
  d[num + 4] = 0x51;
  d[num + 5] = 0x23;
  d[num + 6] = 0x45;
  d[num + 7] = 0x60;
  d[num + 8] = 0x3;
  d[num + 9] = 0x3;
  d[num + 10] = 0x3;
  d[num + 11] = 0x3;

  if (lu->naa) { /* If defined in config file */
    sscanf((const char *)lu->naa,
	   "%hhx:%hhx:%hhx:%hhx:%hhx:%hhx:%hhx:%hhx",
	   &d[num + 4],
	   &d[num + 5],
	   &d[num + 6],
	   &d[num + 7],
	   &d[num + 8],
	   &d[num + 9],
	   &d[num + 10],
	   &d[num + 11]);
  } else { /* Else munge the serial number */
    ptr--;

    for (j = 11; j > 3; ptr--, j--)
      d[num + j] = *ptr;
  }
  /* Bug reported by Stefan Hauser.
   * [num +4] is always 0x5x
   */
  d[num + 4] &= 0x0f;
  d[num + 4] |= 0x50;

  return;
}

#endif

/* Return original slot location if empty
 */
static struct s_info *previous_storage_slot(struct s_info *s,
						struct list_head *slot_head)
{
	struct s_info *sp;	/* Slot Pointer */

	/* Find slot info for 'previous location' */
	list_for_each_entry(sp, slot_head, siblings) {
		if (sp->element_type == STORAGE_ELEMENT)
			if (sp->slot_location == s->last_location)
				if (!slotOccupied(sp))
					/* previous location is empty */
					return sp;
	}

	return NULL;
}

/* Return first empty storage slot.
 */
static struct s_info *find_empty_storage_slot(struct s_info *s,
						struct list_head *slot_head)
{
	struct s_info *sp;	/* Slot Pointer */

	/* If previous location is no good - lets find first empty slot */
	list_for_each_entry(sp, slot_head, siblings) {
		if (!slotOccupied(sp) && sp->element_type == STORAGE_ELEMENT)
			return sp;
	}

	return NULL;
}

/* Save config on shutdown - Not to be called at other times !! */
static void save_config(struct lu_phy_attr *lu)
{
	FILE *ctrl;
	char conf[256];
	struct smc_priv *lu_priv;
	struct list_head *slot_head;
	struct list_head *drive_head;
	struct s_info *sp;	/* Slot Pointer */
	struct d_info *dp;	/* Drive Pointer */
	int last_element_type = 0;

	if (strlen(MHVTL_CONFIG_PATH LIBCONTENTS) >=
				ARRAY_SIZE(conf) - sizeof(".persist")) {
		MHVTL_LOG("Filename length exceeds %d", (int)ARRAY_SIZE(conf));
	}

	snprintf(conf, ARRAY_SIZE(conf), MHVTL_CONFIG_PATH
					LIBCONTENTS "%ld" ".persist", my_id);
	ctrl = fopen(conf, "w");
	if (!ctrl) {
		MHVTL_ERR("Can not open file %s to save state : %s", conf,
					strerror(errno));
		return;
	}

	lu_priv = lu->lu_private;

	drive_head = &lu_priv->drive_list;
	slot_head = &lu_priv->slot_list;

	/* Walk each drive and force-unload into previous location
	 * - if possible */
	list_for_each_entry(dp, drive_head, siblings) {
		if (slotOccupied(dp->slot)) {
			/* Force a move of media from drive into
			 * empty storage slot on shutdown
			 */
			sp = dp->slot;
			MHVTL_DBG(1, "Found %s in drive %d from %d",
					sp->media->barcode,
					sp->slot_location -
						lu_priv->pm->start_drive + 1,
					sp->last_location);
			unload_drive_on_shutdown(sp,
					previous_storage_slot(sp, slot_head));
		}
	}

	/* Walk each drive and force-unload into first empty storage slot */
	list_for_each_entry(dp, drive_head, siblings) {
		if (slotOccupied(dp->slot)) {
			/* Force a move of media from drive into
			 * empty storage slot on shutdown
			 */
			sp = dp->slot;
			MHVTL_DBG(1, "Found %s in drive %d from %d",
					sp->media->barcode,
					sp->slot_location -
						lu_priv->pm->start_drive + 1,
					sp->last_location);
			unload_drive_on_shutdown(sp,
					find_empty_storage_slot(sp, slot_head));
		}
	}

	/* Walk the list of all slots and write data into .persist file */
	list_for_each_entry(sp, slot_head, siblings) {
		/* Pretty up conf file -
		 * Place a blank line between element types
		 */
		if (last_element_type != sp->element_type) {
			last_element_type = sp->element_type;
			fprintf(ctrl, "\n");
		}

		switch (sp->element_type) {
		case DATA_TRANSFER:
			fprintf(ctrl, "Drive %d:\n",
					sp->slot_location -
						lu_priv->pm->start_drive + 1);
			break;
		case MEDIUM_TRANSPORT:
			fprintf(ctrl, "Picker %d: %s\n",
				sp->slot_location -
						lu_priv->pm->start_picker + 1,
				slotOccupied(sp) ? sp->media->barcode : "");
			break;
		case MAP_ELEMENT:
			fprintf(ctrl, "MAP %d: %s\n",
				sp->slot_location -
						lu_priv->pm->start_map + 1,
				slotOccupied(sp) ? sp->media->barcode : "");
			break;
		case STORAGE_ELEMENT:
			fprintf(ctrl, "Slot %d: %s\n",
				sp->slot_location -
						lu_priv->pm->start_storage + 1,
				slotOccupied(sp) ? sp->media->barcode : "");
			break;
		}
	}
	fclose(ctrl);
}

static int init_lu(struct lu_phy_attr *lu, unsigned minor, struct mhvtl_ctl *ctl,
		   char* cont_name)
{
  struct vpd **lu_vpd = lu->lu_vpd;
  int pg;  
  char config[HOME_DIR_PATH_SZ];
  FILE *conf;
  char *b, *obuf, *bp;    /* Read from file into this buffer */
  char *s;	/* Somewhere for sscanf to store results */
  int linecount;
  size_t req_size = 0;
  char buf[16];
  char *v;

  req_size = strlen(conf_directory) + strlen(DEVICE_CONF);
  if (req_size >= sizeof(config)) {
    MHVTL_ERR("Insufficient buffer size detected, buffer size:%lu, "
	      " data size:%lu",	
	      sizeof(config), req_size);
    exit(1);
  }
  snprintf(config, sizeof(config), "%s%s", conf_directory, DEVICE_CONF);
  
  backoff = DEFLT_BACKOFF_VALUE;
  lu->persist = FALSE;

  /* Set static 'home_directory' var - used for get_cart_type() function */
  update_home_dir(my_id, cont_name);
    
  /* Configure default inquiry data */
  memset(&lu->inquiry, 0, MAX_INQUIRY_SZ);
  lu->inquiry[0] = TYPE_MEDIUM_CHANGER;	/* SMC device */
  lu->inquiry[1] = 0x80;	/* Removable bit set */
  /* 
   * For the media changer we are supporting for 
   * our implementation, the ANSI-Approved version
   * needs to be 3 - current ANSI version of the 
   * SCSI-3 spec.
   */
  lu->inquiry[2] = 0x05;	/* SCSI Version (v3) */
  lu->inquiry[3] = 0x02;	/* Response Data Format */
  lu->inquiry[4] = 59;	/* Additional Length */
  lu->inquiry[6] = 0x01;	/* Addr16 */
  lu->inquiry[7] = 0x20;	/* Wbus16 */

  put_unaligned_be16(0x0300, &lu->inquiry[58]); /* SPC-3 No ver claimed */
  put_unaligned_be16(0x0960, &lu->inquiry[60]); /* iSCSI */
  put_unaligned_be16(0x0200, &lu->inquiry[62]); /* SSC */

  lu->ptype = TYPE_MEDIUM_CHANGER;	/* SSC */

  lu->sense_p = &sense[0];	/* Save pointer to sense buffer */

  conf = fopen(config , "r");
  if (!conf) {
    MHVTL_ERR("%s[%s]: Can not open config file[%s] : %s",
	      __FILE__,__func__, config, strerror(errno));
    exit(1);
  }

  s = zalloc(MALLOC_SZ);
  if (!s) {
    perror("Could not allocate memory");
    exit(1);
  }

  obuf = zalloc(8 * MALLOC_SZ);   /* 4K should cover device.conf */
  if (!obuf) {
    MHVTL_ERR("Could not allocate memory (obuf) ...\n");
    exit(1);
  } else {
    /* 
     * Read up everything in one shot and use the new
     * routine, readlinew that works on the buffer. This
     * helps us in tape bring-up time performance.
     * 
     */
    if (fread(obuf, 1, (8 * MALLOC_SZ), conf) < 1) {
      MHVTL_ERR("Could not read device configuration ...\n");
      exit(1);
    }
    bp = obuf;
  }

  lu->fifoname = NULL;
  lu->fifo_fd = NULL;
  lu->fifo_flag = 0;

  smc_slots.movecommand = NULL;
  smc_slots.commandtimeout = 20;

  /* While read in a line */
  linecount = 0;	/* Line count */
  while ((b = readlinew(&bp)) != NULL) {
    linecount++;
    if (b[0] == '#')	/* Ignore comments */
      continue;
    if (!g_serial_base[0] && (sscanf(b, "TAPE_BASE=%s\n", g_serial_base) == 1)) {
      if (strlen(g_serial_base) != 6) {
	MHVTL_LOG("Invalid serial_base %s", g_serial_base);
      }
      continue;
    }
  }
  MHVTL_DBG(1, "g_serial_base: %s", g_serial_base);
  fclose(conf);
  free(obuf);
  free(s);
  
  ctl->channel = 0;
  ctl->id = my_id;
  ctl->lun = 0;
  
  /*
   * Fixed settings for Library - we don't
   * use device.conf any longer. We use 
   * device.conf only for TAPE_BASE
   */
  sprintf(buf, "%s_%.2d", g_serial_base, ctl->lun);
  sprintf(lu->lu_serial_no, "%-10s", replsalt(buf));
  sprintf(lu->vendor_id, "%-8s", is_oem ? oem_vendor_id : VENDOR_ID);
  sprintf(&lu->inquiry[8], "%-8s", is_oem ? oem_vendor_id : VENDOR_ID );
  sprintf(lu->product_id, "%-16s", is_oem ? OEM_PRODUCT_ID : PRODUCT_ID );
  sprintf(&lu->inquiry[16], "%-16s", is_oem ? OEM_PRODUCT_ID : PRODUCT_ID);
  
  lu->naa = zalloc(48);
  if (lu->naa) {
    sprintf((char *)lu->naa, 
	    "%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x",
	    (int)my_id, 0x22, 0x33, 0x44, 
	    0xab, 0, ctl->id, ctl->lun);
  }

  /* Default rev with mhvtl release info */
  v = get_version();
  MHVTL_DBG(1, "Adding default vers info: %s", v);
  sprintf(&lu->inquiry[32], "%-4s", v);
  free(v);

  INIT_LIST_HEAD(&lu->den_list);
  INIT_LIST_HEAD(&lu->log_pg);
  INIT_LIST_HEAD(&lu->mode_pg);
  INIT_LIST_HEAD(&smc_slots.slot_list);
  INIT_LIST_HEAD(&smc_slots.drive_list);
  INIT_LIST_HEAD(&smc_slots.media_list);

  free(smc_slots.state_msg);

  smc_slots.num_drives = 0;
  smc_slots.num_picker = 0;
  smc_slots.num_map = 0;
  smc_slots.num_storage = 0;
  smc_slots.bufsize = SMC_BUF_SIZE;
  smc_slots.state_msg = NULL;

  /* Unit Serial Number */
  pg = PCODE_OFFSET(0x80);
  lu_vpd[pg] = alloc_vpd(strlen(lu->lu_serial_no));
  if (lu_vpd[pg]) {
    update_vpd_80(lu, lu->lu_serial_no);
  } else {
    MHVTL_DBG(1, "Could not malloc(%d) line %d",
	      (int)strlen(lu->lu_serial_no),
	      __LINE__);
  }

  lu->lu_private = &smc_slots;
  smc_slots.cap_closed = CAP_CLOSED;

  return 1;
}

static void process_cmd(int cdev, uint8_t *buf, struct mhvtl_header *mhvtl_cmd,
			useconds_t pollInterval)
{
	struct mhvtl_ds dbuf;
	uint8_t *cdb;

	/* Get the SCSI cdb from vtl driver
	 * - Returns SCSI command S/No. */

	cdb = (uint8_t *)&mhvtl_cmd->cdb;

	/* Interpret the SCSI command & process
	-> Returns no. of bytes to send back to kernel
	 */
	dbuf.sz = 0;
	dbuf.serialNo = mhvtl_cmd->serialNo;
	dbuf.data = buf;
	dbuf.sam_stat = sam_status;
	dbuf.sense_buf = &sense;

	processCommand(cdev, cdb, &dbuf, pollInterval);

	/* Complete SCSI cmd processing */
	completeSCSICommand(cdev, &dbuf);

	/* dbuf.sam_stat was zeroed in completeSCSICommand */
	sam_status = dbuf.sam_stat;
}

/*
 * Be nice and free all malloc() on exit
 */
static void cleanup_lu(struct lu_phy_attr *lu)
{
	int i;
	struct smc_priv *lu_priv;
	struct list_head *slot_head;
	struct s_info *sp, *sn;	/* Slot */
	struct d_info *dp, *dn;	/* Drive */
	struct m_info *mp, *mn;	/* Media */

	lu_priv = lu->lu_private;

	/* Free all VPD pages */
	for (i = 0x80; i < 0x100; i++) {
		if (lu->lu_vpd[PCODE_OFFSET(i)]) {
			dealloc_vpd(lu->lu_vpd[PCODE_OFFSET(i)]);
			lu->lu_vpd[PCODE_OFFSET(i)] = NULL;
		}
	}
	free(lu->naa);
	lu->naa = NULL;

	dealloc_all_mode_pages(lu);
	dealloc_all_log_pages(lu);

	slot_head = &lu_priv->slot_list;
	list_for_each_entry_safe(sp, sn, slot_head, siblings) {
		list_del(&sp->siblings);
		free(sp);
	}

	slot_head = &lu_priv->drive_list;
	list_for_each_entry_safe(dp, dn, slot_head, siblings) {
		list_del(&dp->siblings);
		free(dp);
	}

	slot_head = &lu_priv->media_list;
	list_for_each_entry_safe(mp, mn, slot_head, siblings) {
		list_del(&mp->siblings);
		free(mp);
	}
	free(lu_priv->state_msg);
	lu_priv->state_msg = NULL;
}

static void customise_ibm_lu(struct lu_phy_attr *lu)
{
	if (!strncasecmp(lu->product_id, "3573-TL", 7))
		init_ibmts3100(lu);
	else if (!strncasecmp(lu->product_id, "03584", 5))
		init_ibm3584(lu);
	else
		init_default_smc(lu);
}

static void customise_stk_lu(struct lu_phy_attr *lu)
{
	if (!strncasecmp(lu->product_id, "SL500", 5))
		init_stkslxx(lu);	/* STK SL series */
	else if (!strncasecmp(lu->product_id, "L20", 3))
		init_stkl20(lu);	/* L20/40/80 */
	else if (!strncasecmp(lu->product_id, "L40", 3))
		init_stkl20(lu);	/* L20/40/80 */
	else if (!strncasecmp(lu->product_id, "L80", 3))
		init_stkl20(lu);	/* L20/40/80 */
	else
		init_stklxx(lu);	/* STK L series */
}

static void customise_hp_lu(struct lu_phy_attr *lu)
{
	if (!strncasecmp(lu->product_id, "MSL", 3))
		init_hp_msl_smc(lu);
	else
		init_hp_eml_smc(lu);
}

static void customise_spectra_lu(struct lu_phy_attr *lu)
{
	if (!strncasecmp(lu->product_id, "PYTHON", 6))
		init_spectra_logic_smc(lu);
	else if (!strncasecmp(lu->product_id, "GATOR", 5))
		init_spectra_gator_smc(lu);
	else if (!strncasecmp(lu->product_id, "215", 3))
		init_spectra_215_smc(lu);
	else
		init_spectra_logic_smc(lu);
}

static void customise_lu(struct lu_phy_attr *lu)
{
  if (!strncasecmp(lu->vendor_id, "stk", strlen("stk")) ||
      !strncasecmp(lu->vendor_id, "quest", strlen("quest"))) {
    MHVTL_DBG(1, "%s[customise_stk_lu]: LUN customization of vendor id: %s",
	      __func__, lu->vendor_id);
    customise_stk_lu(lu);
  } else if (!strncasecmp(lu->vendor_id, "IBM", 3)) {
    MHVTL_DBG(1, "%s[customise_ibm_lu]: LUN customization of vendor id: %s",
	      __func__, lu->vendor_id);
    
    customise_ibm_lu(lu);
  } else if (!strncasecmp(lu->vendor_id, "HP", 2)) {
    MHVTL_DBG(1, "%s[customise_hp_lu]: LUN customization of vendor id: %s",
	      __func__, lu->vendor_id);

    customise_hp_lu(lu);
  } else if (!strncasecmp(lu->product_id, "OVERLAND", 8)) {
    MHVTL_DBG(1, "%s[init_overland_smc]: LUN customization of vendor id: %s",
	      __func__, lu->vendor_id);
    
    init_overland_smc(lu);
  } else if (!strncasecmp(lu->product_id, "scalar", 6)) {
    MHVTL_DBG(1, "%s[init_scalar_smc]: LUN customization of vendor id: %s",
	      __func__, lu->vendor_id);
	
    init_scalar_smc(lu);
  } else if (!strncasecmp(lu->vendor_id, "SPECTRA ", 7)) {
    MHVTL_DBG(1, "%s[customise_spectra_lu]: LUN customization of vendor id: %s",
	      __func__, lu->vendor_id);
	
    customise_spectra_lu(lu);
  } else {
    MHVTL_DBG(1, "%s[init_default_smc]: LUN customization of vendor id: %s",
	      __func__, lu->vendor_id);
    
    init_default_smc(lu);
  }
  
  return;
}

void rereadconfig(int sig)
{
	struct mhvtl_ctl ctl;
	unsigned int buffer_size;

	lunit.online = 0;	/* Report library offline until finished */
	current_state = MHVTL_STATE_OFFLINE;
	
	MHVTL_DBG(1, "Caught signal (%d): Re-initialising library %d",
			sig, (int)my_id);

	cleanup_lu(&lunit);

	if (lunit.fifo_fd) {
		fclose(lunit.fifo_fd);
		free(lunit.fifoname);
		lunit.fifo_fd = NULL;
	}

	if (!init_lu(&lunit, my_id, &ctl, cont_name)) {
		fprintf(stderr, "error: Cannot find entry for '%ld' in config file\n",
				my_id);
		exit(1);
	}

	if (lunit.fifoname)
		open_fifo(&lunit.fifo_fd, lunit.fifoname);

	customise_lu(&lunit);

	update_drive_details(&lunit);

	/* malloc a big enough buffer to fit worst case read element status */
	buffer_size = (smc_slots.num_drives +
				smc_slots.num_picker +
				smc_slots.num_map +
				smc_slots.num_storage) * 80;
	reset_device();	/* Force a POWER-ON/RESET sense code */
	if (buffer_size > smc_slots.bufsize) {
		MHVTL_LOG("Too many slots configured"
			" - possible buffer overflow");
		MHVTL_LOG("Please shutdown this daemon and restart so"
			" correct buffer allocation can be performed");
		/* Leave library offline */
	} else {
		lunit.online = 1;	/* Should be good to go */
		current_state = MHVTL_STATE_ONLINE;
	}
}


/*
 *
 * Creates cartridges upon receiving the message .
 * Return 0 on failure, non-zero on sucess
 */
static int create_carts(struct q_msg *msg)
{
	unsigned char sam_stat;
	int pcl, i, rc=0, err = 0;
	unsigned int buffer_size;
	char spcl[10];
	char top_cartdir[10];
	int no_of_tapes;
	char alphanum[4];
	char *msg_argv;
	DIR *d;
	char send_buf[512];
	char async[6];

	/*
	 *  Remove string "create_carts"
	 */
	msg_argv = strtok(msg->text, " ");

	/*
	 * Get num of carts
	 */
	msg_argv = strtok(NULL, " ");
	no_of_tapes = atoi(msg_argv);
	msg_argv = strtok(NULL, " ");

	memset(async, 0, sizeof(async));
	
	if (msg_argv)
	  strncpy(async, msg_argv, sizeof(async)-1);
	
	memset(send_buf, 0, sizeof(send_buf));

	/* 
	 * Validate number of carts requested here 
	 */
	if (gslots >= max_s_slot) {
		if (msg->snd_id != -1 /* -1 = local */) {
			send_msg("FAILED: Max allowed number of cartridges reached, "
				"cannot create anymore", msg->snd_id);
		}
		return 1;
	} else if ((no_of_tapes + gslots) > max_s_slot) {
		no_of_tapes = max_s_slot - gslots;
		sprintf(send_buf, "FAILED: Max allowed cartridges will be reached, "
				"creating only the allowed number of cartridges: ");
		MHVTL_LOG("%s", send_buf);
	} else {
		MHVTL_LOG("Creating %d cartridges", no_of_tapes);
	}
		
	/* Find the top most dir and then monotonically increment */
	memset(top_cartdir, 0, sizeof(top_cartdir));
	memset(spcl, 0, sizeof(spcl));
	
	d = opendir(cart_home);
	if (!d) 
		MHVTL_ERR("cannot open cart home dir");
	
	while(1) {
		struct dirent *entry;
		entry = readdir(d);
		if (!entry){
		  closedir(d);
		  break;
		}

		if (strncmp(entry->d_name, g_serial_base, 6) == 0 &&
		    strcmp(entry->d_name, top_cartdir) > 0) {
		  // Lexicographical order barcode string comparision
		  strncpy(top_cartdir, entry->d_name, sizeof(top_cartdir)-1);
		}
	}

	strncpy(spcl, top_cartdir+6, 3);
	
	pcl = __get_base10(spcl);
	for (i = (pcl+1); i <= (pcl+no_of_tapes); i++){
		__get_base36(i, alphanum);
		snprintf(spcl, sizeof(spcl), "%s%s", g_serial_base, alphanum);
		memset((uint8_t *)&mam, 0, sizeof(mam));
		mam.tape_fmt_version = TAPE_FMT_VERSION;
		mam.mam_fmt_version = MAM_VERSION;
		// tape_size in MBs
		put_unaligned_be64((uint64_t)tape_size * 1048576, &mam.max_capacity); // In bytes
		put_unaligned_be64((uint64_t)tape_size * 1048576, &mam.remaining_capacity); // In bytes
		put_unaligned_be64(sizeof(mam.pad), &mam.MAMSpaceRemaining);
		
		memcpy(&mam.MediumManufacturer, "linuxVTL", 8);
		memcpy(&mam.ApplicationVendor, "vtl-1.7 ", 8);
		sprintf((char *)mam.ApplicationVersion, "%d", TAPE_FMT_VERSION);
		
		mam.MediumType = MEDIA_TYPE_DATA;       /* Normal data cart */
		set_media_params(&mam, density); // "LTO8" - 12 TBs

		snprintf((char *)mam.MediumSerialNumber, sizeof(mam.MediumSerialNumber),
					 "%s_%d", spcl, (int)time(NULL));
		snprintf((char *)mam.MediumManufactureDate, sizeof(mam.MediumManufactureDate), 
					"%d", (int)time(NULL));
		snprintf((char *)mam.Barcode, sizeof(mam.Barcode), "%-31s", spcl);

		/* Create the PCL using the initialized MAM. */
		rc = create_tape(spcl, &mam, &sam_stat, cart_home, &err);

		if (rc) {
			snprintf(send_buf + strlen(send_buf),
					sizeof(send_buf),
					"FAILED: Created cartridges: %d/%d : "
					"%s\n", (i-pcl-1),
					no_of_tapes, strerror(err));
			MHVTL_LOG("sender_id: %ld, %s", msg->snd_id, send_buf);
			if (strcmp(async, "async")) {
				if (msg->snd_id != -1) {
					send_msg(send_buf, msg->snd_id);
				}
			}
			return rc;
		}
		/* init a new storage slot */
		gslots++;
		init_storage_slot(&lunit, gslots, replsalt(spcl), spcl);
	}
        /* malloc a big enough buffer to fit worst case read element status */
        buffer_size = (smc_slots.num_drives +
                                smc_slots.num_picker +
                                smc_slots.num_map +
                                smc_slots.num_storage) * 80;
        init_smc_mode_pages(&lunit);
        init_smc_log_pages(&lunit);
	
	reset_device(); /* power-on reset */
        if (buffer_size > smc_slots.bufsize) {
                MHVTL_LOG("WARNING: Too many slots configured - possible buffer overflow");
                /* Leave library offline */
        }
	buffer_size = max(SMC_BUF_SIZE, (int)buffer_size);
	smc_slots.bufsize = buffer_size;
	snprintf(send_buf + strlen(send_buf), 25, "Created %d cartridges", no_of_tapes);
	if (strcmp(async, "async")) {
		if (msg->snd_id != -1) {
			send_msg(send_buf, msg->snd_id);
		}
	}
	if (msg->snd_id != -1) {
		MHVTL_LOG("sender_id: %ld, %s", msg->snd_id, send_buf);
	} 

	/* New carts created - show */
	push_show = 1;

	/* done */
	return rc;
}

void move_to_pool(struct s_info *sp)
{
	char	export_marker[1024];
	FILE	*fp = NULL;
	int		marker_fd = -1;
	char	pool_marker[1024];
	int		pool_marker_fd = -1;
	/* Very important - who placed it? ROBOT_ARM or OPERATOR? */
	if (sp->status & STATUS_ImpExp) {
		/* 
		 * Operator placed it - so, this is a cart
		 * coming from the cloud - import case. Do nothing.
		 */
		return;
	}

	/* 
	 * Simply move from carts to pool directory 
	 * PM will kick in and start a replication to CSG 
	 */

	/* 
	 * Set the MAP slot empty - meaning it now appears 
	 * as if the customer went to the library, opened
	 * the MAP slot flap and manually took the cartridge
	 * out of the MAP slot ...
	 */
	MHVTL_DBG(3, "Moving [%s] to pool directory...", sp->media->barcode);
	snprintf(export_marker, ARRAY_SIZE(export_marker), "%s/%.9s/export",
			cart_home, sp->media->barcode);
	marker_fd = creat(export_marker, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP);
	if (marker_fd == -1)
	{
		MHVTL_ERR("Failed to create file %s: %s, Cartridge not exported", export_marker,
				strerror(errno));
	} else {
		fsync(marker_fd);
		close(marker_fd);
	}

	snprintf(pool_marker, ARRAY_SIZE(pool_marker), "%s/pool/%.9s",
		   home_directory, sp->media->barcode);

	pool_marker_fd = creat(pool_marker, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP);
	if (pool_marker_fd == -1)
	{
		MHVTL_ERR("Failed to create file %s: %s, Cartridge not exported", pool_marker,
				strerror(errno));
	}else {
		fsync(pool_marker_fd);
		close(pool_marker_fd);
	}

	setSlotEmpty(sp);
	sp->media = NULL;
	sp->last_location = 0;
	push_show = 1;

}

/* Check export slot */
void check_map(void)
{
	struct list_head *slot_head = &smc_slots.slot_list;
	struct s_info *sp;
	int ret = 0;
	ret = restore_cart_from_cloud_done();
	if (ret)
		MHVTL_LOG("restore_cart_from_cloud_done() returned error");
	list_for_each_entry(sp, slot_head, siblings) {
		if (slotOccupied(sp) && sp->element_type == MAP_ELEMENT) {
			/* Move cart file to pool directory */
      			move_to_pool(sp);
    		}
  	}
}

void smc_personality_module_register(struct smc_personality_template *pm)
{
	MHVTL_LOG("%s", pm->name);
	smc_slots.pm = pm;
}

static int cdev;
static struct mhvtl_ctl ctl;
static struct sigaction old_sigint_action;
static int cleaned_up = 0;

static void caught_signal(int signo)
{
  if (signo == SIGINT) {
    if (!cleaned_up++) {
      /* Don't do anything that goes over SMI */
      MHVTL_LOG("SIGINT - cleanup: [%d:%d:%d] ...", 
		ctl.channel, 
		ctl.id, 
		ctl.lun);
      /*
       * With the new approach do not remove
       * LUN - but make it busy so that the devices
       * stay around but always giving out sense
       * data of NOT_READY 0x4, 00 (Cause unknown)
       */
      ioctl(cdev, VTL_BUSY_LU, &ctl);
    }
    signal(signo, SIG_DFL);
    kill(getpid(), signo);
    return;
  } else {
    MHVTL_ERR(" Received signal: %d", signo);
  }
}


int get_tape_size_mb(char *tape_size_gb)
{
  int rval = 0;
  
  if (sscanf(tape_size_gb, "%d", &rval) != 1) {
    return -1;
  }
  
  return (rval * 1024);
}

int
get_max_s_slot(char *limits_table, char *tape_size)
{
  char *saveptr1 = NULL, *saveptr2 = NULL;
  char *tok = NULL;
  char *tape_size_tok = NULL; 
  char *max_s_slot_tok =NULL;
  int rval = 0;
  char *ltable = NULL;

  if (limits_table == NULL) {
    // LTO8 - 12TB, 500 storage slots - 6000TB = 5-6PB(?) per QS VTL
    char *table_map_string = "12TB-500 800GB-1000,400GB-2000,200GB-4000,100GB-5000,50GB-5000,10GB-5000";
    ltable = strdup(table_map_string);
    if (ltable == NULL) {
      MHVTL_ERR("Out of memory allocating memory for limits_table: [\"%s\"]", table_map_string);
      exit(-ENOMEM);
    }
  }

  printf("    max storage slots table [%s]\n", limits_table ? limits_table : ltable);

  for(tok = strtok_r(limits_table ? limits_table : ltable, ",", &saveptr1);
      tok;
      tok = strtok_r(NULL, ",", &saveptr1)) {

    tape_size_tok = strtok_r(tok, "-", &saveptr2);
    max_s_slot_tok = strtok_r(NULL, "-", &saveptr2);

    if (!strncasecmp(tape_size_tok, tape_size,  strlen(tape_size_tok))) {
      if (!sscanf(max_s_slot_tok, "%d", &rval)) {
	printf("    fetching max_s_slot value failed for combination [%s]-[%s]\n",
	       tape_size_tok, max_s_slot_tok);
	if (ltable)
	  free(ltable);
	return -1;
      }
      if (ltable)
	free(ltable);
      return rval;
    }
  }

  if (ltable)
    free(ltable);

  return -1;
}



/*
 * main()
 *
 * e'nuf sed
 */
int main(int argc, char *argv[])
{
	int ret;
	long pollInterval = 0L;
	uint8_t *buf;
	int buffer_size;
	int fifo_retval;
	int last_state = MHVTL_STATE_UNKNOWN;
	struct list_head *slot_head = &smc_slots.slot_list;
	struct s_info *sp;
	struct d_info *dp;
	struct mhvtl_header mhvtl_cmd;
	char s[100];
	pid_t sid, child_cleanup;
	struct sigaction new_action, old_action;

	char *progname = argv[0];
	char *name = "mhvtl";
	char *fifoname = NULL;
	struct passwd *pw;
	struct stat stats;
	char *h_tape_size = NULL;
	int opt;
	int foreground = 0;
	
	/* Message Q */
	int mlen, r_qid;
	struct q_entry r_entry;
	
	memset(&mhvtl_cmd, 0, sizeof(struct mhvtl_header));
	memset(&ctl, 0, sizeof(struct mhvtl_ctl));

	setbuf(stdout, NULL);

	while ((opt = getopt(argc, argv, "adrt:c:s:o:x:v::q:f::F")) != -1) {
		switch (opt) {
		case 'd':
			debug++;
			//verbose = 9;	/* If debug, make verbose... */
			//foreground = 1;
			break;
		case 'v':
			if (optarg)
				verbose = atoi(optarg);
			else
				verbose++;
			/* limit verbosity to single digit */
			if (verbose > 9)
				verbose = 9;
			break;
		case 'r':
			reinstated++;
			break;
		case 'q':
			my_id = atoi(optarg);
			if ((my_id < 0) || (my_id > MAXPRIOR)) {
				fprintf(stderr, "error: queue ID(my_id: %d) out of "
					"range [1..%u]\n", my_id, MAXPRIOR);
				usage(progname);
				exit(1);
			}
			break;
		case 't':
			no_of_drives = atoi(optarg);
			break;
			
		case 'f':
			if (optarg)
				fifoname = strdup(optarg);
			break;
			
		case 'c':
			cont_name = optarg;
			break;
			
		case 's':
			h_tape_size = optarg;
			break;
			
		case 'o':
			is_oem=true;
			oem_vendor_id = optarg;
			break;
			
		case 'x':
			sprintf(g_repl_salt, "%.02d", atoi(optarg));
			break;
			
		case 'a':
			archive = true;
			break;
			
		case 'F':
			foreground = 1;
			break;
		default:
			usage(progname);
			exit(1);
		}
	}

	if (my_id <= 0 || my_id > MAXPRIOR) {
		usage(progname);
		if (my_id == 0)
			printf("    -q must be specified\n");
		else
			printf("    -q value out of range [1 - %d]\n",
				MAXPRIOR);
		exit(1);
	}

	if (no_of_drives <= 0 || no_of_drives > 60) {
		usage(progname);
		if (no_of_drives == 0)
			printf("    -t must be specified\n");
		else
			printf("    -t value out of range [1 - %d]\n",
				60);
		exit(1);
	}


        if (cont_name == NULL) {
                puts("    -c must be specified\n");
                exit(1);
        }
	if (h_tape_size && strlen(h_tape_size) > 0)
	{
	  tape_size = get_tape_size_mb(h_tape_size);
	  if (tape_size < 0) {
            printf("    could not compute the tape size for input [%s]\n",
                   h_tape_size);
            exit(1);
	  }

	  max_s_slot = get_max_s_slot(getenv("OCA_VTL_MAX_TAPES_ASSOC_LIST"), h_tape_size);
	  if (max_s_slot < 0) {
	    printf("    max_s_slot cannot be determined for tape size[%s]\n",
		   h_tape_size);
	    exit(1);
	  }
	} else {
		puts("    -s must be specified\n");
		exit(1);
	}	
 
	GET_CONTAINER_ROOT(home_directory, cont_name);
	if (stat(home_directory, &stats) < 0 && errno == ENOENT) {
                puts("    -c value is not a valid container");
                exit(1);
        }

	GET_CONFIG_PATH(conf_directory, cont_name);

	GET_CART_PATH(cart_home, cont_name);
	       
#ifndef VTL_LOG_WRITER
	openlog(progname, LOG_PID, LOG_DAEMON|LOG_WARNING);
#else
	vtl_log_init();
#endif

	MHVTL_LOG("%s[%s]: container: [%s], config file: [%s], home directory: [%s]\n",
		  __FILE__, __func__, cont_name, conf_directory, home_directory);
	
	/* Disaster Recovery of VTL. Check for drsnap recovery marker file in the container's
	 * root directory.
	 * Example: /containers^/${container_name}/vtl_drsnap_recovery_marker.txt
	 * ^: libocasmi hook needs "containers" prefix in the filesystem paths referring the 
	 * container.
	 */
	{
	  char drsnap_recovery_marker[1024];
	  int rc = 0;
	  struct stat marker_stat;

	  snprintf(drsnap_recovery_marker, ARRAY_SIZE(drsnap_recovery_marker),
		   "%s/vtl_drsnap_recovery_marker.txt",
		   home_directory);
	  
	  if (stat(drsnap_recovery_marker, &marker_stat) == 0) {
	    /*  disaster recovery marker file exists, recover the VTL container */

	    rc = create_missing_carts(drsnap_recovery_marker);
	    if (rc) {
	      MHVTL_ERR("Disaster recovery of VTL conatiner(%s) failed while creating"
			" missing carts", cont_name);
	      /* TODO: have to generate the Qorestor event/alert so that admin can take the
	       * appropriate action.
	       */
	    } else {
	      MHVTL_LOG("Disaster recovery of VTL conatiner(%s) is succesful", cont_name);
	      
	      /* unlink the disaster recovery marker file */
	      unlink(drsnap_recovery_marker);
	    }
	  }

	}

	/* Clear Sense arr */
	memset(sense, 0, sizeof(sense));
	reset_device();	/* power-on reset */

	if (!init_lu(&lunit, my_id, &ctl, cont_name)) {
	  MHVTL_ERR("Can not find entry for '%ld' in config file\n", my_id);
	  vtl_log_close();
	  exit(1);
	}

	/* Personality module init call */
	customise_lu(&lunit);

	update_drive_details(&lunit);

	lunit.online = 1; /* Mark unit online */

	if (chrdev_create(my_id)) {
	  MHVTL_ERR("Error creating device node mhvtl%d", (int)my_id);
	  vtl_log_close();
	  exit(1);
	}

	/* Check for correct user-account before creating lu */
	pw = getpwnam(USR);	/* Find UID for user 'vtl' */
	if (!pw) {
	  MHVTL_ERR("Unable to find user: %s\n", USR);
	  vtl_log_close();
	  exit(1);
	}

	MHVTL_LOG("USR(%s): [%s %s %d %d]", USR, pw->pw_name, pw->pw_gecos, pw->pw_uid,
		  pw->pw_gid);
	
	new_action.sa_handler = caught_signal;
	new_action.sa_flags = 0;
	sigemptyset(&new_action.sa_mask);
	//sigaction(SIGALRM, &new_action, &old_action);
	sigaction(SIGINT, &new_action, &old_sigint_action);
	sigaction(SIGPIPE, &new_action, &old_action);
	sigaction(SIGTERM, &new_action, &old_action);
	sigaction(SIGUSR1, &new_action, &old_action);
	sigaction(SIGUSR2, &new_action, &old_action);

	new_action.sa_handler = rereadconfig;
	sigaction(SIGHUP, &new_action, &old_action);

	child_cleanup = add_lu(my_id, &ctl);
	INIT_LIST_HEAD(&import_serial_list.list);
	if (!child_cleanup) {
	  MHVTL_ERR("Could not add logical unit [ ID : %ld ]", my_id);
	  vtl_log_close();
	  exit(1);
	}

	chrdev_chown(my_id, pw->pw_uid, pw->pw_gid);

	/* Now that we have created the lu, drop root uid/gid */
	if (setgid(pw->pw_gid)) {
		MHVTL_ERR("Unable to change gid: %d", errno);
		vtl_log_close();
		exit(1);
	}
	if (setuid(pw->pw_uid)) {
		MHVTL_ERR("Unable to change uid: %d", errno);
		vtl_log_close();
		exit(1);
	}

	MHVTL_DBG(2, "Running as %s, uid: %d", pw->pw_name, getuid());
	
	/* Initialise message queue as necessary */
	r_qid = init_queue();
	if (r_qid == -1) {
	  MHVTL_ERR("Could not initialise message queue: %d\n", errno);
	  vtl_log_close();
	  exit(1);
	}

	if (check_for_running_daemons(my_id)) {
	  MHVTL_LOG("%s: version %s, found another running daemon... "
		    "exiting", progname, MHVTL_VERSION);
	  vtl_log_close();
	  exit(2);
	}

	/* Clear out message Q by reading anything there. */
	mlen = msgrcv(r_qid, &r_entry, MAXOBN, my_id, IPC_NOWAIT);
	while (mlen > 0) {
		MHVTL_DBG(2, "Found \"%s\" still in message Q", r_entry.msg.text);
		mlen = msgrcv(r_qid, &r_entry, MAXOBN, my_id, IPC_NOWAIT);
	}

	cdev = chrdev_open(name, my_id);
	if (cdev == -1) {
		MHVTL_ERR("Could not open /dev/%s%ld: %s",
					name, my_id, strerror(errno));
		fflush(NULL);
		vtl_log_close();
		exit(1);
	}

	/* malloc a big enough buffer to fit worst case read element status */
	buffer_size = (smc_slots.num_drives +
				smc_slots.num_picker +
				smc_slots.num_map +
				smc_slots.num_storage) * 80;
	buffer_size = max(SMC_BUF_SIZE, buffer_size);
	smc_slots.bufsize = buffer_size;
	MHVTL_DBG(1, "Setting buffer size to %d", buffer_size);
	buf = (uint8_t *)zalloc(buffer_size);
	if (NULL == buf) {
	  MHVTL_ERR("Problems allocating memory: %d", errno);
	  vtl_log_close();
	  exit(1);
	}

	/* Send a message to each tape drive so they know the
	 * controlling library's message Q id
	 */
	list_for_each_entry(sp, slot_head, siblings) {
		if (sp->element_type == DATA_TRANSFER) {
			dp = sp->drive;
			if (send_msg("Register", dp->drv_id) < 0) {
			  MHVTL_DBG(3, "Unable to send msg: \"Register\" to drive with "
				    "id: %d", dp->drv_id);
			} else {
			  MHVTL_DBG(3, "Sent msg: \"Register\" to drive with id: %d",
				    dp->drv_id);

			}

			if (debug) {

				MHVTL_DBG(3, "\nDrive %d", sp->slot_location);

				strncpy(s, dp->inq_vendor_id, 8+2);
				rmnl(s, ' ', 8);
				s[8] = '\0';
				MHVTL_DBG(3, "Vendor ID     : \"%s\"", s);

				strncpy(s, dp->inq_product_id, 16+2);
				rmnl(s, ' ', 16);
				s[16] = '\0';
				MHVTL_DBG(3, "Product ID    : \"%s\"", s);

				strncpy(s, dp->inq_product_rev, 4+2);
				rmnl(s, ' ', 4);
				s[4] = '\0';
				MHVTL_DBG(3, "Revision Level: \"%s\"", s);

				strncpy(s, dp->inq_product_sno, 10+2);
				rmnl(s, ' ', 10);
				s[10] = '\0';
				MHVTL_DBG(3, "Product S/No  : \"%s\"", s);

				MHVTL_DBG(3, "Drive location: %d",
						dp->slot->slot_location);
				MHVTL_DBG(3, "Drive occupied: %s",
				(dp->slot->status & STATUS_Full) ? "No" : "Yes");
			}
		}
	}

	/*
	 * We cannot daemonize for this implementation since 
	 * that affects reclamation of events in SMI because 
	 * SMI can get the pid of the parent but not of the
	 * child process which runs the for loop below. 
	 */

	umask(0);	/* Change the file mode mask */
	
	sid = setsid();
	if (sid < 0) {
	  vtl_log_close();
	  exit(-1);
	}

	if ((chdir(home_directory)) < 0) {
	  MHVTL_ERR("Unable to change directory to home_directory");
	  vtl_log_close();
	  exit(-1);
	}
	
	close(STDIN_FILENO);
	close(STDERR_FILENO);

	MHVTL_LOG("Started %s: version %s, verbose log lvl: %d, lu [%d:%d:%d]",
					progname, MHVTL_VERSION, verbose,
					ctl.channel, ctl.id, ctl.lun);

	oom_adjust();

	/* Strict support for STK L700 */
	if (STK) {
		/* size of dvcid area in RES descriptor */
		smc_slots.dvcid_len = 32;

		/* dvcid area only contains a serial number */
		smc_slots.dvcid_serial_only = TRUE;
	} else if (!strcmp(lunit.vendor_id, "SPECTRA ") &&
		/* Tweak as necessary to properly mimic the specified library type. */
		/* FIXME: This moves into personality modules once implemented */
		!strcmp(lunit.product_id, "PYTHON          ")) {
		/* size of dvcid area in RES descriptor */
		smc_slots.dvcid_len = 10;
		/* dvcid area only contains a serial number */
		smc_slots.dvcid_serial_only = TRUE;
	} else {
		/* size of dvcid area in RES descriptor */
		smc_slots.dvcid_len = 34;
		/* dvcid area contains vendor, product, serial */
		smc_slots.dvcid_serial_only = FALSE;
	}

	/* Reference Quantum 6-00423013 SCSI Reference - Rev A */
	if (!strncmp(lunit.product_id, "Scalar", 6)) {
		int h, m, sec;
		int day, month, year;

		month = 0;

		MHVTL_DBG(3, "Detected Scalar library");

		ymd(&year, &month, &day, &h, &m, &sec);

		/* Controller firmware build date */
		sprintf((char *)&lunit.inquiry[36], "%04d-%02d-%02d %02d:%02d:%02d",
				year, month, day, h, m, sec);

		lunit.inquiry[55] = 0x01;	/* Contains barcode scanner : BarC */
	}
	
	/* If fifoname passed as switch */
	if (fifoname)
		process_fifoname(&lunit, fifoname, 1);
	
	/* fifoname can be defined in device.conf */
	if (lunit.fifoname) {
		open_fifo(&lunit.fifo_fd, lunit.fifoname);
	}

	fifo_retval = inc_fifo_count();
	if (fifo_retval == -ENOMEM) {
		MHVTL_ERR("shared memory setup failed - exiting...");
		goto exit;
	} else if (fifo_retval < 0) {
		MHVTL_ERR("Failed to set fifo count()...");
	}


	/* 
	 * Unlink the info file 
	 */
	sprintf(g_f, "/tmp/info.%ld", my_id);
	unlink(g_f);

 	/*
 	 * Send a message to all tape drives if we
	 * have been reinstated ...
	 */
	if (reinstated) {
		unload_td(&lunit, -1);
	}

	current_state = MHVTL_STATE_ONLINE;
	/*
	 * Main loop for the vtllibrary. 
	 */
	for (;;) {
		/* We cleaned up elsewhere */
		if (cleaned_up)
		  goto exit;

		/* 
		 * Auto-create cartridges since the user 
		 * created a VTL but he doesn't have any
		 * cartridges to play with
		 */
		if (!child_cleanup && !gslots) {
			struct q_msg q;

			MHVTL_LOG("Auto-creating 60 cartridges ...");
			q.snd_id = -1;
			memset(q.text, 0, sizeof(q.text));
			sprintf(q.text, "create_carts 60 async");
			ioctl(cdev, VTL_BUSY_LU, &ctl);
			create_carts(&q);
			ioctl(cdev, VTL_UNBUSY_LU, &ctl);
		}
		
		/*
		 * Push vtl --stats to /tmp
		 */
		if (push_show)
		  vtl_show(&lunit);

		/* 
		 * Check for any messages from the vtllibrary or 
		 * from management scripts via vtlcmd tool.
		 */
		mlen = msgrcv(r_qid, &r_entry, MAXOBN, my_id, IPC_NOWAIT);
		if (mlen > 0) {
			/*
			 * Got an "exit" message
			 */
			if (child_cleanup) {
				MHVTL_DBG(1, "vtllibrary is initializing, skipping any "
					  "request");
			} else if (processMessageQ(&r_entry.msg, &ctl, cdev)) {
			  // processMessageQ returns 1 if it got "exit" message. so exit
				goto exit;
			}
		} else if (mlen < 0) {
			r_qid = init_queue();
			if (r_qid == -1)
				MHVTL_ERR("Can not open message queue: %s",
					strerror(errno));
		}

		/*
		 * Ask the pseudo HBA driver if it has anything for us.
		 */
		ret = ioctl(cdev, VTL_POLL_AND_GET_HEADER, &mhvtl_cmd);
		if (ret < 0) {
			MHVTL_LOG("ret: %d : %s", ret, strerror(errno));
		} else {
			if (child_cleanup) {
				if (waitpid(child_cleanup, NULL, WNOHANG)) {
					MHVTL_DBG(2,
						"Cleaning up after add_lu "
						"child pid: %d",
							child_cleanup);
					child_cleanup = 0;
				}
			}
			fflush(NULL);   /* So I can pipe debug o/p thru tee */
			switch (ret) {
			  /*
			   * Pseudo HBA gave us a CDB to process
			   */
			case VTL_QUEUE_CMD:
				if (smc_slots.bufsize != buffer_size) {
					buffer_size = smc_slots.bufsize;
					buf = realloc(buf, buffer_size);
					if (!buf) {
						perror("Problems allocating memory");
						exit(1);
					}
				}
				process_cmd(cdev, buf, &mhvtl_cmd, pollInterval);
				pollInterval = MIN_SLEEP_TIME;
				break;
				
				/*
				 * There was nothing in the queue - increase
				 * time between ioctls for performance reasons
				 */
			case VTL_IDLE:
				usleep(pollInterval);

				if (pollInterval < 1000000)
					pollInterval += backoff;
				break;

				/*
				 * Return is not expected
				 */
			default:
				/* MHVTL_LOG("ioctl(0x%x) returned %d", */
				/* 		VTL_POLL_AND_GET_HEADER, ret); */
				sleep(1);
				break;
			}

			if (current_state != last_state) {
				status_change(lunit.fifo_fd,
							current_state,
							my_id,
							&smc_slots.state_msg);
				
				last_state = current_state;
			}
			
			if (pollInterval > 0x18000) {
			  if (current_state != MHVTL_STATE_OFFLINE)
			    current_state = MHVTL_STATE_IDLE;
			}
			
			/* Check on MAP slots */
			if (!(smc_slots.iec++ % 1000)) {
			  check_map();
			}
		}

		if (lunit.cp & 0x1) {
			MHVTL_LOG("Executing crash point # 1 ..."); 
			sleep(1);
			assert(0);
		}
	}
exit:
	if (cleaned_up)
	  exit (0);
	
	cleaned_up++;
	MHVTL_LOG("Exiting %s: [%d:%d:%d] ...", 
				progname, 
				ctl.channel, 
				ctl.id, 
				ctl.lun);

	/*
	 * LUN - but make it busy so that the devices
	 * stay around but always giving out sense
	 * data of NOT_READY 0x4, 00 (Cause unknown)
	 */
	ioctl(cdev, VTL_BUSY_LU, &ctl);

	cleanup_lu(&lunit);

	if (lunit.persist)
		save_config(&lunit);

	close(cdev);
	free(buf);
	dec_fifo_count();
	if (lunit.fifo_fd) {
		fclose(lunit.fifo_fd);
		unlink(lunit.fifoname);
		free(lunit.fifoname);
	}

	vtl_log_close();
	/* Close open handle and then delete the file */
	if (g_fp)
	  fclose(g_fp);
	if (g_f[0])
	  unlink(g_f);
	
	exit(0);
}
